# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""This module provides tools for running workflows in a streaming fashion."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from copy import deepcopy
from typing import Any, Generic, TypeVar

import networkx as nx
import sciline
import scipp as sc

T = TypeVar('T')


def maybe_hist(value: T) -> T:
    """
    Convert value to a histogram if it is not already a histogram.

    This is the default pre-processing used by accumulators.

    Parameters
    ----------
    value:
        Value to be converted to a histogram.

    Returns
    -------
    :
        Histogram.
    """
    if not isinstance(value, sc.Variable | sc.DataArray):
        return value
    return value if value.bins is None else value.hist()


class Accumulator(ABC, Generic[T]):
    """
    Abstract base class for accumulators.

    Accumulators are used to accumulate values over multiple chunks.
    """

    def __init__(self, preprocess: Callable[[T], T] | None = maybe_hist) -> None:
        """
        Parameters
        ----------
        preprocess:
            Preprocessing function to be applied to pushed values prior to accumulation.
        """
        self._preprocess = preprocess

    def push(self, value: T) -> None:
        """
        Push a value to the accumulator.

        Parameters
        ----------
        value:
            Value to be pushed to the accumulator.
        """
        if self._preprocess is not None:
            value = self._preprocess(value)
        self._do_push(value)

    @abstractmethod
    def _do_push(self, value: T) -> None: ...

    @property
    def is_empty(self) -> bool:
        """
        Check if the accumulator is empty.

        Returns
        -------
        :
            True if the accumulator is empty, False otherwise.
        """
        return False

    @property
    def value(self) -> T:
        """
        Get the accumulated value.

        Returns
        -------
        :
            Accumulated value.

        Raises
        ------
        ValueError
            If the accumulator is empty.
        """
        if self.is_empty:
            raise ValueError("Cannot get value from empty accumulator")
        return self._get_value()

    @abstractmethod
    def _get_value(self) -> T:
        """Return the accumulated value, assuming it exists."""

    @abstractmethod
    def clear(self) -> None:
        """
        Clear the accumulator, resetting it to its initial state.
        """


class EternalAccumulator(Accumulator[T]):
    """
    Simple accumulator that adds pushed values immediately.

    Does not support event data.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._value: T | None = None

    @property
    def is_empty(self) -> bool:
        return self._value is None

    def _get_value(self) -> T:
        return deepcopy(self._value)

    def _do_push(self, value: T) -> None:
        if self._value is None:
            self._value = deepcopy(value)
        else:
            self._value += value

    def clear(self) -> None:
        """Clear the accumulated value."""
        self._value = None


class MeanAccumulator(EternalAccumulator[T]):
    """
    Accumulator that computes the mean of pushed values.

    Does not support event data.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._count = 0

    def _do_push(self, value: T) -> None:
        super()._do_push(value)
        self._count += 1

    def _get_value(self) -> T:
        return super()._get_value() / self._count

    def clear(self) -> None:
        """Clear the accumulated value and count."""
        super().clear()
        self._count = 0


class RollingAccumulator(Accumulator[T]):
    """
    Accumulator that adds pushed values to a rolling window.

    Does not support event data.
    """

    def __init__(self, window: int = 10, **kwargs: Any) -> None:
        """
        Parameters
        ----------
        window:
            Size of the rolling window.
        """
        super().__init__(**kwargs)
        self._window = window
        self._values: list[T] = []

    @property
    def is_empty(self) -> bool:
        return len(self._values) == 0

    def _get_value(self) -> T:
        # Naive and potentially slow implementation if values and/or window are large!
        return sc.reduce(self._values).sum()

    def _do_push(self, value: T) -> None:
        self._values.append(value)
        if len(self._values) > self._window:
            self._values.pop(0)

    def clear(self) -> None:
        """Clear the accumulated values."""
        self._values = []


class MinAccumulator(Accumulator):
    """Keeps the minimum value seen so far.

    Only supports scalar values.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._cur_min: sc.Variable | None = None

    def _do_push(self, value: sc.Variable) -> None:
        if self._cur_min is None:
            self._cur_min = value
        else:
            self._cur_min = min(self._cur_min, value)

    @property
    def is_empty(self) -> bool:
        """Check if the accumulator has collected a minimum value."""
        return self._cur_min is None

    def _get_value(self) -> Any:
        return self._cur_min

    def clear(self) -> None:
        """Clear the accumulated minimum value."""
        self._cur_min = None


class MaxAccumulator(Accumulator):
    """Keeps the maximum value seen so far.

    Only supports scalar values.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._cur_max: sc.Variable | None = None

    @property
    def is_empty(self) -> bool:
        """Check if the accumulator has collected a maximum value."""
        return self._cur_max is None

    def _do_push(self, value: sc.Variable) -> None:
        if self._cur_max is None:
            self._cur_max = value
        else:
            self._cur_max = max(self._cur_max, value)

    def _get_value(self) -> sc.Variable | None:
        return self._cur_max

    def clear(self) -> None:
        """Clear the accumulated maximum value."""
        self._cur_max = None


class StreamProcessor:
    """
    Wrap a base workflow for streaming processing of chunks.

    Note that this class can not determine if the workflow is valid for streamed
    processing based on the input keys. In particular, it is the responsibility of the
    user to ensure that the workflow is "linear" with respect to the dynamic keys up to
    the accumulation keys.

    Similarly, the stream processor cannot determine from the workflow structure whether
    context updates are compatible with the accumulated data. Accumulators are not
    cleared automatically. This is best illustrated with an example:

    - If the context is the detector rotation angle, and we accumulate I(Q) (or a
      prerequisite of I(Q)), then updating the detector angle context is compatible with
      previous data, assuming Q for each new chunk is computed based on the angle.
    - If the context is the sample temperature, and we accumulate I(Q), then updating
      the temperature context is not compatible with previous data. Accumulating I(Q, T)
      could be compatible in this case.

    Since the correctness cannot be determined from the workflow structure, we recommend
    implementing processing steps in a way to catch such problems. For example, adding
    the temperature as a coordinate to the I(Q) data array should allow for
    automatically raising in the accumulator if the temperature changes.
    """

    def __init__(
        self,
        base_workflow: sciline.Pipeline,
        *,
        dynamic_keys: tuple[sciline.typing.Key, ...],
        context_keys: tuple[sciline.typing.Key, ...] = (),
        target_keys: tuple[sciline.typing.Key, ...],
        accumulators: dict[sciline.typing.Key, Accumulator | Callable[..., Accumulator]]
        | tuple[sciline.typing.Key, ...],
        allow_bypass: bool = False,
    ) -> None:
        """
        Create a stream processor.

        Parameters
        ----------
        base_workflow:
            Workflow to be used for processing chunks.
        dynamic_keys:
            Keys that are expected to be updated with each chunk. These keys cannot
            depend on each other or on context_keys.
        context_keys:
            Keys that define context for processing chunks and may change occasionally.
            These keys cannot overlap with dynamic_keys or depend on each other or on
            dynamic_keys.
        target_keys:
            Keys to be computed and returned.
        accumulators:
            Keys at which to accumulate values and their accumulators. If a tuple is
            passed, :py:class:`EternalAccumulator` is used for all keys. Otherwise, a
            dict mapping keys to accumulator instances can be passed. If a dict value is
            a callable, base_workflow.bind_and_call(value) is used to make an instance.
        allow_bypass:
            If True, allow bypassing accumulators for keys that are not in the
            accumulators dict. This is useful for dynamic keys that are not "terminated"
            in any accumulator. USE WITH CARE! This will lead to incorrect results
            unless the values for these keys are valid for all chunks comprised in the
            final accumulators at the point where :py:meth:`finalize` is called.
        """
        self._dynamic_keys = set(dynamic_keys)
        self._context_keys = set(context_keys)

        # Validate that dynamic and context keys do not overlap
        overlap = self._dynamic_keys & self._context_keys
        if overlap:
            raise ValueError(f"Keys cannot be both dynamic and context: {overlap}")

        # Check dynamic/context keys don't depend on other dynamic/context keys
        graph = base_workflow.underlying_graph
        special_keys = self._dynamic_keys | self._context_keys
        for key in special_keys:
            if key not in graph:
                continue
            ancestors = nx.ancestors(graph, key)
            special_ancestors = ancestors & special_keys
            downstream = 'Dynamic' if key in self._dynamic_keys else 'Context'
            if special_ancestors:
                raise ValueError(
                    f"{downstream} key '{key}' depends on other dynamic/context keys: "
                    f"{special_ancestors}. This is not supported."
                )

        workflow = sciline.Pipeline()
        for key in target_keys:
            workflow[key] = base_workflow[key]
        for key in dynamic_keys:
            workflow[key] = None  # hack to prune branches
        for key in context_keys:
            workflow[key] = None

        # Find and pre-compute static nodes as far down the graph as possible
        nodes = _find_descendants(workflow, dynamic_keys + context_keys)
        last_static = _find_parents(workflow, nodes) - nodes
        for key, value in base_workflow.compute(last_static).items():
            workflow[key] = value

        # Nodes that may need updating on context change but should be cached otherwise.
        dynamic_nodes = _find_descendants(workflow, dynamic_keys)
        # Nodes as far "down" in the graph as possible, right before the dynamic nodes.
        # This also includes target keys that are not dynamic but context-dependent.
        context_to_cache = (
            (_find_parents(workflow, dynamic_nodes) | set(target_keys)) - dynamic_nodes
        ) & _find_descendants(workflow, context_keys)
        graph = workflow.underlying_graph
        self._context_key_to_cached_context_nodes_map = {
            context_key: ({context_key} | nx.descendants(graph, context_key))
            & context_to_cache
            for context_key in self._context_keys
            if context_key in graph
        }

        self._context_workflow = workflow.copy()
        self._process_chunk_workflow = workflow.copy()
        self._finalize_workflow = workflow.copy()
        self._accumulators = (
            accumulators
            if isinstance(accumulators, dict)
            else {key: EternalAccumulator() for key in accumulators}
        )

        # Map each accumulator to its dependent dynamic keys
        self._accumulator_dependencies = {
            acc_key: nx.ancestors(graph, acc_key) & self._dynamic_keys
            for acc_key in self._accumulators
            if acc_key in graph
        }

        # Depending on the target_keys, some accumulators can be unused and should not
        # be computed when adding a chunk.
        self._accumulators = {
            key: value for key, value in self._accumulators.items() if key in graph
        }
        # Create accumulators unless instances were passed. This allows for initializing
        # accumulators with arguments that depend on the workflow such as bin edges,
        # which would otherwise be hard to obtain.
        self._accumulators = {
            key: value
            if isinstance(value, Accumulator)
            else base_workflow.bind_and_call(value)
            for key, value in self._accumulators.items()
        }
        self._target_keys = target_keys
        self._allow_bypass = allow_bypass

    def set_context(self, context: dict[sciline.typing.Key, Any]) -> None:
        """
        Set the context for processing chunks.

        Parameters
        ----------
        context:
            Context to be set.
        """
        needs_recompute = set()
        for key in context:
            if key not in self._context_keys:
                raise ValueError(f"Key '{key}' is not a context key")
            needs_recompute |= self._context_key_to_cached_context_nodes_map[key]
        for key, value in context.items():
            self._context_workflow[key] = value
        results = self._context_workflow.compute(needs_recompute)
        for key, value in results.items():
            if key in self._target_keys:
                # Context-dependent key is direct target, independent of dynamic nodes.
                self._finalize_workflow[key] = value
            else:
                self._process_chunk_workflow[key] = value

    def add_chunk(
        self, chunks: dict[sciline.typing.Key, Any]
    ) -> dict[sciline.typing.Key, Any]:
        """
        Legacy interface for accumulating values from chunks and finalizing the result.

        It is recommended to use :py:meth:`accumulate` and :py:meth:`finalize` instead.

        Parameters
        ----------
        chunks:
            Chunks to be processed.

        Returns
        -------
        :
            Finalized result.
        """
        self.accumulate(chunks)
        return self.finalize()

    def accumulate(self, chunks: dict[sciline.typing.Key, Any]) -> None:
        """
        Accumulate values from chunks without finalizing the result.

        Parameters
        ----------
        chunks:
            Chunks to be processed.

        Raises
        ------
        ValueError
            If non-dynamic keys are provided in chunks.
            If accumulator computation requires dynamic keys not provided in chunks.
        """
        non_dynamic = set(chunks) - self._dynamic_keys
        if non_dynamic:
            raise ValueError(
                f"Can only update dynamic keys. Got non-dynamic keys: {non_dynamic}"
            )

        accumulators_to_update = []
        for acc_key, deps in self._accumulator_dependencies.items():
            if deps.isdisjoint(chunks.keys()):
                continue
            if not deps.issubset(chunks.keys()):
                raise ValueError(
                    f"Accumulator '{acc_key}' requires dynamic keys "
                    f"{deps - chunks.keys()} not provided in the current chunk."
                )
            accumulators_to_update.append(acc_key)

        for key, value in chunks.items():
            self._process_chunk_workflow[key] = value
            # There can be dynamic keys that do not "terminate" in any accumulator. In
            # that case, we need to make sure they can be and are used when computing
            # the target keys.
            if self._allow_bypass:
                self._finalize_workflow[key] = value
        to_accumulate = self._process_chunk_workflow.compute(accumulators_to_update)
        for key, processed in to_accumulate.items():
            self._accumulators[key].push(processed)

    def finalize(self) -> dict[sciline.typing.Key, Any]:
        """
        Get the final result by computing the target keys based on accumulated values.

        Returns
        -------
        :
            Finalized result.
        """
        for key in self._accumulators:
            self._finalize_workflow[key] = self._accumulators[key].value
        return self._finalize_workflow.compute(self._target_keys)

    def clear(self) -> None:
        """
        Clear all accumulators, resetting them to their initial state.

        This is useful for restarting a streaming computation without
        creating a new StreamProcessor instance.
        """
        for accumulator in self._accumulators.values():
            accumulator.clear()


def _find_descendants(
    workflow: sciline.Pipeline, keys: tuple[sciline.typing.Key, ...]
) -> set[sciline.typing.Key]:
    graph = workflow.underlying_graph
    descendants = set()
    for key in keys:
        descendants |= nx.descendants(graph, key)
    return descendants | set(keys)


def _find_parents(
    workflow: sciline.Pipeline, keys: tuple[sciline.typing.Key, ...]
) -> set[sciline.typing.Key]:
    graph = workflow.underlying_graph
    parents = set()
    for key in keys:
        parents |= set(graph.predecessors(key))
    return parents



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc
from scippneutron.chopper import DiskChopper
from scippneutron.conversion.graph.beamline import beamline as beamline_graph
from scippneutron.conversion.graph.tof import elastic as elastic_graph

from ess.reduce import time_of_flight
from ess.reduce.nexus.types import DetectorData, SampleRun
from ess.reduce.time_of_flight import GenericTofWorkflow, TofLookupTableWorkflow, fakes

sl = pytest.importorskip("sciline")


def dream_choppers() -> dict[str, DiskChopper]:
    psc1 = DiskChopper(
        frequency=sc.scalar(14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(286 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, -70.405], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=[-1.23, 70.49, 84.765, 113.565, 170.29, 271.635, 286.035, 301.17],
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=[1.23, 73.51, 88.035, 116.835, 175.31, 275.565, 289.965, 303.63],
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    psc2 = DiskChopper(
        frequency=sc.scalar(-14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-236, unit="deg"),
        axle_position=sc.vector(value=[0, 0, -70.395], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=[-1.23, 27.0, 55.8, 142.385, 156.765, 214.115, 257.23, 315.49],
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=[1.23, 30.6, 59.4, 145.615, 160.035, 217.885, 261.17, 318.11],
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    oc = DiskChopper(
        frequency=sc.scalar(14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(297 - 180 - 90, unit="deg"),
        axle_position=sc.vector(value=[0, 0, -70.376], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-27.6 * 0.5], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[27.6 * 0.5], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    bcc = DiskChopper(
        frequency=sc.scalar(112.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(240 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, -66.77], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-36.875, 143.125], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[36.875, 216.875], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    t0 = DiskChopper(
        frequency=sc.scalar(28.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(280 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, -63.5], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-314.9 * 0.5], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[314.9 * 0.5], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    return {"psc1": psc1, "psc2": psc2, "oc": oc, "bcc": bcc, "t0": t0}


def dream_choppers_with_frame_overlap() -> dict[str, DiskChopper]:
    out = dream_choppers()
    out["bcc"] = DiskChopper(
        frequency=sc.scalar(112.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(240 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, -66.77], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-36.875, 143.125], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[56.875, 216.875], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )
    return out


def dream_source_position() -> sc.Variable:
    return sc.vector(value=[0, 0, -76.55], unit="m")


@pytest.fixture(scope="module")
def lut_workflow_dream_choppers() -> sl.Pipeline:
    lut_wf = TofLookupTableWorkflow()
    lut_wf[time_of_flight.DiskChoppers] = dream_choppers()
    lut_wf[time_of_flight.SourcePosition] = dream_source_position()
    lut_wf[time_of_flight.NumberOfSimulatedNeutrons] = 100_000
    lut_wf[time_of_flight.SimulationSeed] = 432
    lut_wf[time_of_flight.PulseStride] = 1
    lut_wf[time_of_flight.SimulationResults] = lut_wf.compute(
        time_of_flight.SimulationResults
    )
    return lut_wf


def setup_workflow(
    raw_data: sc.DataArray,
    ltotal: sc.Variable,
    lut_workflow: sl.Pipeline,
    error_threshold: float = 0.1,
) -> sl.Pipeline:
    pl = GenericTofWorkflow(run_types=[SampleRun], monitor_types=[])
    pl[DetectorData[SampleRun]] = raw_data
    pl[time_of_flight.DetectorLtotal[SampleRun]] = ltotal

    lut_wf = lut_workflow.copy()
    lut_wf[time_of_flight.LtotalRange] = ltotal.min(), ltotal.max()
    lut_wf[time_of_flight.LookupTableRelativeErrorThreshold] = error_threshold

    pl[time_of_flight.TimeOfFlightLookupTable] = lut_wf.compute(
        time_of_flight.TimeOfFlightLookupTable
    )
    return pl


@pytest.mark.parametrize(
    "ltotal",
    [
        sc.array(dims=["detector_number"], values=[77.675], unit="m"),
        sc.array(dims=["detector_number"], values=[77.675, 76.5], unit="m"),
        sc.array(
            dims=["y", "x"],
            values=[[77.675, 76.1, 78.05], [77.15, 77.3, 77.675]],
            unit="m",
        ),
    ],
)
@pytest.mark.parametrize("time_offset_unit", ["s", "ms", "us", "ns"])
@pytest.mark.parametrize("distance_unit", ["m", "mm"])
def test_dream_wfm(
    lut_workflow_dream_choppers, ltotal, time_offset_unit, distance_unit
):
    monitors = {
        f"detector{i}": ltot for i, ltot in enumerate(ltotal.flatten(to="detector"))
    }

    # Create some neutron events
    beamline = fakes.FakeBeamline(
        choppers=dream_choppers(),
        monitors=monitors,
        run_length=sc.scalar(1 / 14, unit="s") * 4,
        events_per_pulse=10_000,
        seed=77,
        source_position=dream_source_position(),
    )

    raw = sc.concat(
        [beamline.get_monitor(key)[0].squeeze() for key in monitors.keys()],
        dim="detector",
    ).fold(dim="detector", sizes=ltotal.sizes)

    # Convert the time offset to the unit requested by the test
    raw.bins.coords["event_time_offset"] = raw.bins.coords["event_time_offset"].to(
        unit=time_offset_unit, copy=False
    )
    # Convert the distance to the unit requested by the test
    raw.coords["Ltotal"] = raw.coords["Ltotal"].to(unit=distance_unit, copy=False)

    # Save reference data
    ref = beamline.get_monitor(next(iter(monitors)))[1].squeeze()
    ref = sc.sort(ref, key='id')

    pl = setup_workflow(
        raw_data=raw, ltotal=ltotal, lut_workflow=lut_workflow_dream_choppers
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph)

    for da in wavs.flatten(to='pixel'):
        x = sc.sort(da.value, key='id')
        diff = abs(
            (x.coords["wavelength"] - ref.coords["wavelength"])
            / ref.coords["wavelength"]
        )
        assert np.nanpercentile(diff.values, 100) < 0.02
        assert sc.isclose(ref.data.sum(), da.data.sum(), rtol=sc.scalar(1.0e-3))


@pytest.fixture(scope="module")
def lut_workflow_dream_choppers_time_overlap():
    lut_wf = TofLookupTableWorkflow()
    lut_wf[time_of_flight.DiskChoppers] = dream_choppers_with_frame_overlap()
    lut_wf[time_of_flight.SourcePosition] = dream_source_position()
    lut_wf[time_of_flight.NumberOfSimulatedNeutrons] = 100_000
    lut_wf[time_of_flight.SimulationSeed] = 432
    lut_wf[time_of_flight.PulseStride] = 1
    lut_wf[time_of_flight.SimulationResults] = lut_wf.compute(
        time_of_flight.SimulationResults
    )
    return lut_wf


@pytest.mark.parametrize(
    "ltotal",
    [
        sc.array(dims=["detector_number"], values=[77.675], unit="m"),
        sc.array(dims=["detector_number"], values=[77.675, 76.5], unit="m"),
        sc.array(
            dims=["y", "x"],
            values=[[77.675, 76.1, 78.05], [77.15, 77.3, 77.675]],
            unit="m",
        ),
    ],
)
@pytest.mark.parametrize("time_offset_unit", ["s", "ms", "us", "ns"])
@pytest.mark.parametrize("distance_unit", ["m", "mm"])
def test_dream_wfm_with_subframe_time_overlap(
    lut_workflow_dream_choppers_time_overlap,
    ltotal,
    time_offset_unit,
    distance_unit,
):
    monitors = {
        f"detector{i}": ltot for i, ltot in enumerate(ltotal.flatten(to="detector"))
    }

    # Create some neutron events
    beamline = fakes.FakeBeamline(
        choppers=dream_choppers_with_frame_overlap(),
        monitors=monitors,
        run_length=sc.scalar(1 / 14, unit="s") * 4,
        events_per_pulse=10_000,
        seed=88,
        source_position=dream_source_position(),
    )

    raw = sc.concat(
        [beamline.get_monitor(key)[0].squeeze() for key in monitors.keys()],
        dim="detector",
    ).fold(dim="detector", sizes=ltotal.sizes)

    # Convert the time offset to the unit requested by the test
    raw.bins.coords["event_time_offset"] = raw.bins.coords["event_time_offset"].to(
        unit=time_offset_unit, copy=False
    )
    # Convert the distance to the unit requested by the test
    raw.coords["Ltotal"] = raw.coords["Ltotal"].to(unit=distance_unit, copy=False)

    # Save reference data
    ref = beamline.get_monitor(next(iter(monitors)))[1].squeeze()
    ref = sc.sort(ref, key='id')

    pl = setup_workflow(
        raw_data=raw,
        ltotal=ltotal,
        lut_workflow=lut_workflow_dream_choppers_time_overlap,
        error_threshold=0.01,
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph)

    for da in wavs.flatten(to='pixel'):
        x = sc.sort(da.value, key='id')
        sel = sc.isfinite(x.coords["wavelength"])
        y = ref.coords["wavelength"][sel]
        diff = abs((x.coords["wavelength"][sel] - y) / y)
        assert np.nanpercentile(diff.values, 100) < 0.02
        sum_wfm = da.hist(wavelength=100).data.sum()
        sum_ref = ref.hist(wavelength=100).data.sum()
        # Verify that we lost some neutrons that were in the overlapping region
        assert sum_wfm < sum_ref
        assert sum_wfm > sum_ref * 0.9


def v20_choppers():
    wfm1 = DiskChopper(
        frequency=sc.scalar(-70.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-47.10, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 6.6], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=np.array([83.71, 140.49, 193.26, 242.32, 287.91, 330.3]) + 15.0,
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=np.array([94.7, 155.79, 212.56, 265.33, 314.37, 360.0]) + 15.0,
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    wfm2 = DiskChopper(
        frequency=sc.scalar(-70.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-76.76, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 7.1], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=np.array([65.04, 126.1, 182.88, 235.67, 284.73, 330.32]) + 15.0,
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=np.array([76.03, 141.4, 202.18, 254.97, 307.74, 360.0]) + 15.0,
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    foc1 = DiskChopper(
        frequency=sc.scalar(-56.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-62.40, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 8.8], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=np.array([74.6, 139.6, 194.3, 245.3, 294.8, 347.2]),
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=np.array([95.2, 162.8, 216.1, 263.1, 310.5, 371.6]),
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    foc2 = DiskChopper(
        frequency=sc.scalar(-28.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-12.27, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 15.9], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=np.array([98.0, 154.0, 206.8, 255.0, 299.0, 344.65]),
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=np.array([134.6, 190.06, 237.01, 280.88, 323.56, 373.76]),
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    pol = DiskChopper(
        frequency=sc.scalar(-14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(0.0, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 17.0], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=np.array([40.0]),
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=np.array([240.0]),
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )
    return {"wfm1": wfm1, "wfm2": wfm2, "foc1": foc1, "foc2": foc2, "pol": pol}


def v20_source_position():
    return sc.vector([0, 0, 0], unit='m')


@pytest.fixture(scope="module")
def lut_workflow_v20_choppers():
    lut_wf = TofLookupTableWorkflow()
    lut_wf[time_of_flight.DiskChoppers] = v20_choppers()
    lut_wf[time_of_flight.SourcePosition] = v20_source_position()
    lut_wf[time_of_flight.NumberOfSimulatedNeutrons] = 300_000
    lut_wf[time_of_flight.SimulationSeed] = 431
    lut_wf[time_of_flight.PulseStride] = 1
    lut_wf[time_of_flight.SimulationResults] = lut_wf.compute(
        time_of_flight.SimulationResults
    )
    return lut_wf


@pytest.mark.parametrize(
    "ltotal",
    [
        sc.array(dims=["detector_number"], values=[26.0], unit="m"),
        sc.array(dims=["detector_number"], values=[26.0, 25.5], unit="m"),
        sc.array(
            dims=["y", "x"], values=[[26.0, 25.1, 26.33], [25.9, 26.0, 25.7]], unit="m"
        ),
    ],
)
@pytest.mark.parametrize("time_offset_unit", ["s", "ms", "us", "ns"])
@pytest.mark.parametrize("distance_unit", ["m", "mm"])
def test_v20_compute_wavelengths_from_wfm(
    lut_workflow_v20_choppers, ltotal, time_offset_unit, distance_unit
):
    monitors = {
        f"detector{i}": ltot for i, ltot in enumerate(ltotal.flatten(to="detector"))
    }

    # Create some neutron events
    beamline = fakes.FakeBeamline(
        choppers=v20_choppers(),
        monitors=monitors,
        run_length=sc.scalar(1 / 14, unit="s") * 4,
        events_per_pulse=10_000,
        seed=99,
    )

    raw = sc.concat(
        [beamline.get_monitor(key)[0].squeeze() for key in monitors.keys()],
        dim="detector",
    ).fold(dim="detector", sizes=ltotal.sizes)

    # Convert the time offset to the unit requested by the test
    raw.bins.coords["event_time_offset"] = raw.bins.coords["event_time_offset"].to(
        unit=time_offset_unit, copy=False
    )
    # Convert the distance to the unit requested by the test
    raw.coords["Ltotal"] = raw.coords["Ltotal"].to(unit=distance_unit, copy=False)

    # Save reference data
    ref = beamline.get_monitor(next(iter(monitors)))[1].squeeze()
    ref = sc.sort(ref, key='id')

    pl = setup_workflow(
        raw_data=raw, ltotal=ltotal, lut_workflow=lut_workflow_v20_choppers
    )

    tofs = pl.compute(time_of_flight.DetectorTofData[SampleRun])

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph)

    for da in wavs.flatten(to='pixel'):
        x = sc.sort(da.value, key='id')
        diff = abs(
            (x.coords["wavelength"] - ref.coords["wavelength"])
            / ref.coords["wavelength"]
        )
        assert np.nanpercentile(diff.values, 99) < 0.02
        assert sc.isclose(ref.data.sum(), da.data.sum(), rtol=sc.scalar(1.0e-3))



"""Domain types for use with Sciline, parametrized by run- and monitor-type."""

from dataclasses import dataclass
from pathlib import Path
from typing import Any, BinaryIO, Generic, NewType, TypeVar

import sciline
import scipp as sc
import scippnexus as snx
from scippneutron import metadata as scn_meta
from scippneutron.chopper import DiskChopper

FilePath = NewType('FilePath', Path)
"""Full path to a NeXus file on disk."""
NeXusFile = NewType('NeXusFile', BinaryIO)
"""An open NeXus file.

Can be any file handle for reading binary data.

Note that this cannot be used as a parameter in Sciline as there are no
concrete implementations of ``BinaryIO``.
The type alias is provided for callers of load functions outside of pipelines.
"""
NeXusGroup = NewType('NeXusGroup', snx.Group)
"""A ScippNexus group in an open file."""

NeXusEntryName = NewType('NeXusEntryName', str)
"""Name of an entry in a NeXus file."""
NeXusSourceName = NewType('NeXusSourceName', str)
"""Name of a source in a NeXus file."""

DetectorBankSizes = NewType("DetectorBankSizes", dict[str, dict[str, int | Any]])

GravityVector = NewType('GravityVector', sc.Variable)

PreopenNeXusFile = NewType('PreopenNeXusFile', bool)
"""Whether to preopen NeXus files before passing them to the rest of the workflow."""


# 1  TypeVars used to parametrize the generic parts of the workflow

# 1.1  Run types
BackgroundRun = NewType('BackgroundRun', int)
"""Background run such as a run with only a solvent which the sample is placed in."""
EmptyBeamRun = NewType('EmptyBeamRun', int)
"""
Run with empty sample holder, sometimes called 'direct run'.

It is used for reading the data from the transmission monitor.
"""
SampleRun = NewType('SampleRun', int)
"""Sample run."""
VanadiumRun = NewType('VanadiumRun', int)
"""Vanadium run."""

ScatteringRunType = TypeVar(
    'ScatteringRunType',
    BackgroundRun,
    SampleRun,
    VanadiumRun,
)


class TransmissionRun(Generic[ScatteringRunType]):
    """
    Mapping between ScatteringRunType and transmission run.

    In the case where no transmission run is provided, the transmission run should be
    the same as the measurement (sample or background) run.
    """


RunType = TypeVar('RunType')
"""TypeVar for specifying what run some data belongs to.

This type must be constrained when used in a Sciline pipeline.
E.g., by passing ``run_types`` to :class:`ess.reduce.nexus.GenericNeXusWorkflow`.

ESSreduce provides the following but custom types can be used:

- :class:`BackgroundRun`
- :class:`EmptyBeamRun`
- :class:`SampleRun`
- :class:`TransmissionRun`
- :class:`VanadiumRun`
"""


# 1.2  Monitor types
IncidentMonitor = NewType('IncidentMonitor', int)
"""Incident monitor"""
TransmissionMonitor = NewType('TransmissionMonitor', int)
"""Transmission monitor"""
FrameMonitor0 = NewType('FrameMonitor0', int)
"""Frame monitor number 0"""
FrameMonitor1 = NewType('FrameMonitor1', int)
"""Frame monitor number 1"""
FrameMonitor2 = NewType('FrameMonitor2', int)
"""Frame monitor number 2"""
FrameMonitor3 = NewType('FrameMonitor3', int)
"""Frame monitor number 3"""
CaveMonitor = NewType('CaveMonitor', int)
"""A monitor located in the instrument cave"""
MonitorType = TypeVar('MonitorType')
"""TypeVar for specifying what monitor some data belongs to.

This type must be constrained when used in a Sciline pipeline.
E.g., by passing ``monitor_types`` to :class:`ess.reduce.nexus.GenericNeXusWorkflow`.

ESSreduce provides the following but custom types can be used:

- :class:`IncidentMonitor`
- :class:`TransmissionMonitor`
- :class:`FrameMonitor0`
- :class:`FrameMonitor1`
- :class:`FrameMonitor2`
- :class:`FrameMonitor3`
- :class:`CaveMonitor`
"""


Component = TypeVar('Component')
"""A beamline component in a neXus file."""
COMPONENT_CONSTRAINTS = (
    snx.NXdetector,
    snx.NXsample,
    snx.NXsource,
    snx.NXdisk_chopper,
    snx.NXcrystal,
)
"""Base constraints for the Component type variable.

This list will be supplemented with monitor types when creating a pipeline.
"""

UniqueComponent = TypeVar('UniqueComponent', snx.NXsample, snx.NXsource)
"""Components that can be identified by their type as there will only be one."""

Beamline = scn_meta.Beamline
"""Beamline metadata."""
Measurement = scn_meta.Measurement
"""measurement metadata."""
Source = scn_meta.Source
"""Neutron source metadata."""


class NeXusName(sciline.Scope[Component, str], str):
    """Name of a component in a NeXus file."""


class NeXusClass(sciline.Scope[Component, type], type):
    """NX_class of a component in a NeXus file."""


NeXusDetectorName = NeXusName[snx.NXdetector]
"""Name of a detector (bank) in a NeXus file."""


class NeXusComponent(sciline.Scope[Component, RunType, sc.DataGroup], sc.DataGroup):
    """Raw data from a NeXus component."""


class AllNeXusComponents(sciline.Scope[Component, RunType, sc.DataGroup], sc.DataGroup):
    """Raw data from all NeXus components of one class."""


class NeXusData(sciline.Scope[Component, RunType, sc.DataArray], sc.DataArray):
    """
    Data array loaded from an NXevent_data or NXdata group.

    This must be contained in an NXmonitor or NXdetector group.
    """


class Position(sciline.Scope[Component, RunType, sc.Variable], sc.Variable):
    """Position of a component such as source, sample, monitor, or detector."""


class DetectorPositionOffset(sciline.Scope[RunType, sc.Variable], sc.Variable):
    """Offset for the detector position, added to base position."""


class MonitorPositionOffset(
    sciline.Scope[RunType, MonitorType, sc.Variable], sc.Variable
):
    """Offset for the monitor position, added to base position."""


class CalibratedDetector(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Calibrated data from a detector."""


class CalibratedBeamline(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Calibrated beamline with detector and other components."""


class CalibratedMonitor(
    sciline.Scope[RunType, MonitorType, sc.DataArray], sc.DataArray
):
    """Calibrated data from a monitor."""


class DetectorData(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Calibrated detector merged with neutron event or histogram data."""


class MonitorData(sciline.Scope[RunType, MonitorType, sc.DataArray], sc.DataArray):
    """Calibrated monitor merged with neutron event or histogram data."""


class Filename(sciline.Scope[RunType, Path], Path): ...


@dataclass
class TimeInterval(Generic[RunType]):
    """Range of neutron pulses to load from NXevent_data or NXdata groups."""

    value: slice


@dataclass
class NeXusFileSpec(Generic[RunType]):
    value: FilePath | NeXusFile | NeXusGroup


@dataclass
class NeXusLocationSpec:
    """
    NeXus filename and optional parameters to identify (parts of) a component to load.
    """

    filename: FilePath | NeXusFile | NeXusGroup
    entry_name: NeXusEntryName | None = None
    component_name: str | None = None
    selection: snx.typing.ScippIndex | slice = ()


@dataclass
class NeXusComponentLocationSpec(NeXusLocationSpec, Generic[Component, RunType]):
    """
    NeXus filename and optional parameters to identify (parts of) a component to load.
    """


@dataclass
class NeXusAllLocationSpec:
    """
    NeXus parameters to identify all components of a class to load.
    """

    filename: FilePath | NeXusFile | NeXusGroup
    entry_name: NeXusEntryName | None = None
    selection: snx.typing.ScippIndex | slice = ()


@dataclass
class NeXusAllComponentLocationSpec(NeXusAllLocationSpec, Generic[Component, RunType]):
    """
    NeXus parameters to identify all components of a class to load.
    """


@dataclass
class NeXusDataLocationSpec(NeXusLocationSpec, Generic[Component, RunType]):
    """NeXus filename and parameters to identify (parts of) detector data to load."""


class NeXusTransformationChain(
    sciline.Scope[Component, RunType, snx.TransformationChain],
    snx.TransformationChain,
): ...


@dataclass
class NeXusTransformation(Generic[Component, RunType]):
    value: sc.Variable

    @staticmethod
    def from_chain(
        chain: NeXusTransformationChain[Component, RunType],
    ) -> 'NeXusTransformation[Component, RunType]':
        """
        Convert a transformation chain to a single transformation.

        As transformation chains may be time-dependent, this method will need to select
        a specific time point to convert to a single transformation. This may include
        averaging as well as threshold checks. This is not implemented yet and we
        therefore currently raise an error if the transformation chain does not compute
        to a scalar.
        """
        if chain.transformations.sizes != {}:
            raise ValueError(f"Expected scalar transformation, got {chain}")
        transform = chain.compute()
        return NeXusTransformation(value=transform)


class RawChoppers(
    sciline.Scope[RunType, sc.DataGroup[sc.DataGroup[Any]]],
    sc.DataGroup[sc.DataGroup[Any]],
):
    """All choppers in a NeXus file."""


class DiskChoppers(
    sciline.Scope[RunType, sc.DataGroup[DiskChopper]],
    sc.DataGroup[DiskChopper],
):
    """All disk choppers parsed from a NeXus file."""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import sciline
import scipp as sc
import scippnexus as snx
from scipp.testing import assert_identical

from ess.reduce import time_of_flight
from ess.reduce.nexus.types import (
    CalibratedBeamline,
    DetectorData,
    DiskChoppers,
    NeXusData,
    SampleRun,
)
from ess.reduce.time_of_flight import GenericTofWorkflow, TofLookupTableWorkflow, fakes

sl = pytest.importorskip("sciline")


@pytest.fixture
def calibrated_beamline():
    return sc.DataArray(
        data=sc.ones(dims=["detector_number"], shape=[10]),
        coords={
            "Ltotal": sc.scalar(80.0, unit="m"),
            "detector_number": sc.array(
                dims=["detector_number"], values=np.arange(10), unit=None
            ),
        },
    )


@pytest.fixture
def nexus_data():
    events = sc.DataArray(
        data=sc.ones(dims=["event"], shape=[1000]),
        coords={
            "event_time_offset": sc.linspace(
                "event", 0.0, 1000.0 / 14, num=1000, unit="ms"
            ).to(unit="ns"),
            "event_id": sc.array(
                dims=["event"], values=np.arange(1000) % 10, unit=None
            ),
        },
    )
    return sc.DataArray(
        sc.bins(
            begin=sc.array(dims=["pulse"], values=[0], unit=None),
            data=events,
            dim="event",
        )
    )


def test_TofLookupTableWorkflow_can_compute_tof_lut():
    wf = TofLookupTableWorkflow()
    wf[DiskChoppers] = fakes.psc_choppers()
    wf[time_of_flight.NumberOfSimulatedNeutrons] = 10_000
    wf[time_of_flight.LtotalRange] = (
        sc.scalar(0.0, unit="m"),
        sc.scalar(100.0, unit="m"),
    )
    wf[time_of_flight.SourcePosition] = fakes.source_position()
    lut = wf.compute(time_of_flight.TimeOfFlightLookupTable)
    assert isinstance(lut, sc.DataArray)


def test_GenericTofWorkflow_with_tof_lut_from_tof_simulation(
    calibrated_beamline: sc.DataArray, nexus_data: sc.DataArray
):
    wf = GenericTofWorkflow(run_types=[SampleRun], monitor_types=[])
    wf[CalibratedBeamline[SampleRun]] = calibrated_beamline
    wf[NeXusData[snx.NXdetector, SampleRun]] = nexus_data

    # Should be able to compute DetectorData without chopper and simulation params
    # This contains event_time_offset (time-of-arrival).
    _ = wf.compute(DetectorData[SampleRun])
    # By default, the workflow tries to load the LUT from file
    with pytest.raises(sciline.UnsatisfiedRequirement):
        _ = wf.compute(time_of_flight.TimeOfFlightLookupTable)
    with pytest.raises(sciline.UnsatisfiedRequirement):
        _ = wf.compute(time_of_flight.DetectorTofData[SampleRun])

    lut_wf = TofLookupTableWorkflow()
    lut_wf[DiskChoppers] = fakes.psc_choppers()
    lut_wf[time_of_flight.NumberOfSimulatedNeutrons] = 10_000
    lut_wf[time_of_flight.LtotalRange] = (
        sc.scalar(0.0, unit="m"),
        sc.scalar(100.0, unit="m"),
    )
    lut_wf[time_of_flight.SourcePosition] = fakes.source_position()
    table = lut_wf.compute(time_of_flight.TimeOfFlightLookupTable)

    wf[time_of_flight.TimeOfFlightLookupTable] = table
    # Should now be able to compute DetectorData with chopper and simulation params
    detector = wf.compute(time_of_flight.DetectorTofData[SampleRun])
    assert 'tof' in detector.bins.coords


def test_GenericTofWorkflow_with_tof_lut_from_file(
    calibrated_beamline: sc.DataArray,
    nexus_data: sc.DataArray,
    tmp_path: pytest.TempPathFactory,
):
    lut_wf = TofLookupTableWorkflow()
    lut_wf[DiskChoppers] = fakes.psc_choppers()
    lut_wf[time_of_flight.NumberOfSimulatedNeutrons] = 10_000
    lut_wf[time_of_flight.LtotalRange] = (
        sc.scalar(0.0, unit="m"),
        sc.scalar(100.0, unit="m"),
    )
    lut_wf[time_of_flight.SourcePosition] = fakes.source_position()
    lut = lut_wf.compute(time_of_flight.TimeOfFlightLookupTable)
    lut.save_hdf5(filename=tmp_path / "lut.h5")

    wf = GenericTofWorkflow(run_types=[SampleRun], monitor_types=[])
    wf[CalibratedBeamline[SampleRun]] = calibrated_beamline
    wf[NeXusData[snx.NXdetector, SampleRun]] = nexus_data
    wf[time_of_flight.TimeOfFlightLookupTableFilename] = (
        tmp_path / "lut.h5"
    ).as_posix()

    loaded_lut = wf.compute(time_of_flight.TimeOfFlightLookupTable)
    assert_identical(lut, loaded_lut)

    detector = wf.compute(time_of_flight.DetectorTofData[SampleRun])
    assert 'tof' in detector.bins.coords



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc

from ess.reduce import streaming


def test_eternal_accumulator_sums_everything() -> None:
    accum = streaming.EternalAccumulator()
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(10):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, sc.sum(var))


def test_eternal_accumulator_sums_everything_with_preprocess() -> None:
    accum = streaming.EternalAccumulator(preprocess=lambda x: x**0.5)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(10):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, sc.sum(var**0.5))


def test_eternal_accumulator_works_if_output_value_is_modified() -> None:
    accum = streaming.EternalAccumulator()
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(10):
        accum.push(var[i].copy())
    value = accum.value
    value += 1.0
    assert sc.identical(accum.value, sc.sum(var))


def test_eternal_accumulator_does_not_modify_pushed_values() -> None:
    accum = streaming.EternalAccumulator()
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    original = var.copy()
    for i in range(10):
        accum.push(var[i])
    assert sc.identical(var, original)


def test_eternal_accumulator_clear() -> None:
    accum = streaming.EternalAccumulator()
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(10):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, sc.sum(var))
    accum.clear()
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value


def test_rolling_accumulator_sums_over_window() -> None:
    accum = streaming.RollingAccumulator(window=3)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    accum.push(var[0].copy())
    assert sc.identical(accum.value, var[0])
    accum.push(var[1].copy())
    assert sc.identical(accum.value, var[0:2].sum())
    accum.push(var[2].copy())
    assert sc.identical(accum.value, var[0:3].sum())
    accum.push(var[3].copy())
    assert sc.identical(accum.value, var[1:4].sum())
    accum.push(var[4].copy())
    assert sc.identical(accum.value, var[2:5].sum())


def test_rolling_accumulator_sums_over_window_with_preprocess() -> None:
    accum = streaming.RollingAccumulator(window=3, preprocess=lambda x: x**0.5)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    accum.push(var[0].copy())
    assert sc.identical(accum.value, var[0] ** 0.5)
    accum.push(var[1].copy())
    assert sc.identical(accum.value, (var[0:2] ** 0.5).sum())
    accum.push(var[2].copy())
    assert sc.identical(accum.value, (var[0:3] ** 0.5).sum())
    accum.push(var[3].copy())
    assert sc.identical(accum.value, (var[1:4] ** 0.5).sum())
    accum.push(var[4].copy())
    assert sc.identical(accum.value, (var[2:5] ** 0.5).sum())


def test_rolling_accumulator_works_if_output_value_is_modified() -> None:
    accum = streaming.RollingAccumulator(window=3)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(10):
        accum.push(var[i].copy())
    value = accum.value
    value += 1.0
    assert sc.identical(accum.value, var[7:10].sum())


def test_rolling_accumulator_does_not_modify_pushed_values() -> None:
    accum = streaming.RollingAccumulator(window=3)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    original = var.copy()
    for i in range(10):
        accum.push(var[i])
    assert sc.identical(var, original)


def test_rolling_accumulator_clear() -> None:
    accum = streaming.RollingAccumulator(window=3)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(5):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, var[2:5].sum())
    accum.clear()
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value


def test_eternal_accumulator_is_empty() -> None:
    accum = streaming.EternalAccumulator()
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value

    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    accum.push(var[0].copy())
    assert not accum.is_empty
    assert sc.identical(accum.value, var[0])

    accum.clear()
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value


def test_rolling_accumulator_is_empty() -> None:
    accum = streaming.RollingAccumulator(window=3)
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value

    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    accum.push(var[0].copy())
    assert not accum.is_empty
    assert sc.identical(accum.value, var[0])

    accum.clear()
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value


def test_min_accumulator() -> None:
    accum = streaming.MinAccumulator()
    var = sc.array(dims=['x'], values=[1.0, 2.0, 3.0, 2.0, 1.0])
    for scalar_var in var:
        accum.push(scalar_var)
    assert sc.identical(accum.value, sc.min(var))
    accum.clear()
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value


@pytest.mark.parametrize("Accum", [streaming.MinAccumulator, streaming.MaxAccumulator])
def test_accumulator_non_scalar_raises(Accum) -> None:
    accum = Accum()
    var = sc.array(dims=['x'], values=[1.0, 2.0, 3.0, 2.0, 1.0])
    accum.push(var)  # First push does not raise
    with pytest.raises(sc.DimensionError, match="Expected 0 dimensions"):
        accum.push(var)


def test_max_accumulator() -> None:
    accum = streaming.MaxAccumulator()
    var = sc.array(dims=['x'], values=[1.0, 2.0, 3.0, 2.0, 1.0])
    for scalar_var in var:
        accum.push(scalar_var)
    assert sc.identical(accum.value, sc.max(var))
    accum.clear()
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value


def test_mean_accumulator_calculates_mean() -> None:
    accum = streaming.MeanAccumulator()
    var = sc.linspace(dim='x', start=0, stop=10, num=5)
    for i in range(5):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, sc.mean(var))


def test_mean_accumulator_with_preprocess() -> None:
    accum = streaming.MeanAccumulator(preprocess=lambda x: x**2)
    var = sc.linspace(dim='x', start=0, stop=10, num=5)
    for i in range(5):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, sc.mean(var**2))


def test_mean_accumulator_works_if_output_value_is_modified() -> None:
    accum = streaming.MeanAccumulator()
    var = sc.linspace(dim='x', start=0, stop=10, num=5)
    for i in range(5):
        accum.push(var[i].copy())
    value = accum.value
    value += 1.0
    assert sc.identical(accum.value, sc.mean(var))


def test_mean_accumulator_does_not_modify_pushed_values() -> None:
    accum = streaming.MeanAccumulator()
    var = sc.linspace(dim='x', start=0, stop=10, num=5)
    original = var.copy()
    for i in range(5):
        accum.push(var[i])
    assert sc.identical(var, original)


def test_mean_accumulator_clear() -> None:
    accum = streaming.MeanAccumulator()
    var = sc.linspace(dim='x', start=0, stop=10, num=5)
    for i in range(5):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, sc.mean(var))
    accum.clear()
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value


def test_mean_accumulator_is_empty() -> None:
    accum = streaming.MeanAccumulator()
    assert accum.is_empty
    with pytest.raises(ValueError, match="Cannot get value from empty accumulator"):
        _ = accum.value

    var = sc.linspace(dim='x', start=0, stop=10, num=5)
    accum.push(var[0].copy())
    assert not accum.is_empty
    assert sc.identical(accum.value, var[0])


def test_mean_accumulator_incremental_mean_calculation() -> None:
    accum = streaming.MeanAccumulator()
    var = sc.array(dims=['x'], values=[2.0, 4.0, 6.0, 8.0])

    # Push values one by one and verify mean is calculated correctly each time
    accum.push(var[0])
    assert sc.identical(accum.value, var[0])

    accum.push(var[1])
    assert sc.identical(accum.value, sc.mean(var[0:2]))

    accum.push(var[2])
    assert sc.identical(accum.value, sc.mean(var[0:3]))

    accum.push(var[3])
    assert sc.identical(accum.value, sc.mean(var))


def test_mean_accumulator_after_clear() -> None:
    accum = streaming.MeanAccumulator()
    var = sc.array(dims=['x'], values=[2.0, 4.0, 6.0])

    # First round of accumulation
    for i in range(3):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, sc.mean(var))

    # Clear and verify empty state
    accum.clear()
    assert accum.is_empty

    # Second round with new values
    new_var = sc.array(dims=['x'], values=[10.0, 20.0])
    for i in range(2):
        accum.push(new_var[i].copy())

    # Verify mean is calculated correctly with only the new values
    assert sc.identical(accum.value, sc.mean(new_var))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""
Raw count processing and visualization for live data display.

The main feature of this module if :py:class:`RollingDetectorView`, which provides a
rolling view of a detector. This is useful for live data display, where the most recent
counts, i.e., from the last N pulses, are of interest.

As various detector geometries can often not be directly visualized in a useful way,
the module also provides a way to project the detector data onto a different coordinate
systems. A full set of options is not implemented yet. Currently there are three
options:

- `'xy_plane'`: Project the data onto the x-y plane, i.e., perpendicular to the beam.
- `'cylinder_mantle_z'`: Project the data onto the mantle of a cylinder aligned with the
   z-axis.
- A callable, e.g., to select and flatten dimensions of the data.
"""

from __future__ import annotations

from collections.abc import Callable, Sequence
from math import ceil
from typing import Literal, NewType

import numpy as np
import scipp as sc
import scippnexus as snx

from ess.reduce.nexus.types import (
    CalibratedDetector,
    Filename,
    NeXusComponent,
    NeXusDetectorName,
    NeXusTransformation,
    SampleRun,
)
from ess.reduce.nexus.workflow import GenericNeXusWorkflow

from . import roi

CalibratedPositionWithNoisyReplicas = NewType(
    'CalibratedPositionWithNoisyReplicas', sc.Variable
)
DetectorViewResolution = NewType('DetectorViewResolution', dict[str, int])
PixelCylinderAxis = NewType('PixelCylinderAxis', sc.Variable)
PixelCylinderRadius = NewType('PixelCylinderRadius', sc.Variable)
PixelShape = NewType('PixelShape', sc.DataGroup)
PositionNoise = NewType('PositionNoise', sc.Variable)
PositionNoiseReplicaCount = NewType('PositionNoiseReplicaCount', int)
PositionNoiseSigma = NewType('PositionNoiseSigma', sc.Variable)
ProjectedCoords = NewType('ProjectedCoords', sc.DataGroup[str, sc.Variable])
RollingDetectorViewWindow = NewType('RollingDetectorViewWindow', int)


class Histogrammer:
    """
    Histogrammer for a detector view.

    The histogrammer is used to project the detector data onto a different coordinate
    system. This class implements the common logic for different projections. For each
    concrete projection a setup mechanism computes relevant coordinates. This is done,
    e.g., in :py:func:`project_xy` and :py:func:`project_onto_cylinder`.
    """

    def __init__(
        self,
        coords: sc.DataGroup[str, sc.Variable],
        edges: sc.DataGroup[str, sc.Variable],
    ):
        self._current = 0
        self._replica_dim = 'replica'
        self._replicas = coords.sizes[self._replica_dim]
        self._coords = coords
        self._edges = edges

    @property
    def replicas(self) -> int:
        return self._replicas

    @staticmethod
    def from_coords(
        coords: ProjectedCoords, resolution: DetectorViewResolution
    ) -> Histogrammer:
        """
        Create a histogrammer from coordinates and resolution.

        Parameters
        ----------
        coords:
            Coordinates to use for histogramming. May contain more than the
            dimensions used for histogramming.
        resolution:
            Resolution to use for histogramming. The keys are the dimensions
            to histogram, and the values are the number of bins to use. The order
            of the dimensions is preserved in the output and thus controls which axis
            is which in a plot.
        """
        edges = sc.DataGroup(
            {
                dim: coords[dim].hist({dim: res}).coords[dim]
                for dim, res in resolution.items()
            }
        )
        return Histogrammer(coords=coords, edges=edges)

    def __call__(self, da: sc.DataArray) -> sc.DataArray:
        self._current += 1
        coords = self._coords[self._replica_dim, self._current % self._replicas]
        return self._hist(da.data, coords=coords)

    def _hist(self, data: sc.Variable, *, coords: sc.DataGroup) -> sc.DataArray:
        # If input is multi-dim we need to flatten since those dims cannot be preserved.
        return sc.DataArray(data, coords=coords).flatten(to='_').hist(self._edges)

    def input_indices(self) -> sc.DataArray:
        """Return an array with input indices corresponding to each histogram bin."""
        dim = 'detector_number'
        # For some projections one of the coords is a scalar, convert to flat table.
        coords = self._coords.broadcast(sizes=self._coords.sizes).flatten(to=dim).copy()
        ndet = sc.index(coords.sizes[dim] // self._replicas)
        da = sc.DataArray(
            sc.arange(dim, coords.sizes[dim], dtype='int64', unit=None) % ndet,
            coords=coords,
        )
        return sc.DataArray(da.bin(self._edges).bins.data, coords=self._edges)

    def apply_full(self, var: sc.Variable) -> sc.DataArray:
        """
        Apply the histogrammer to a variable using all replicas.

        This is used for one-off operations where the full data is needed, e.g., for
        transforming pixel weights. Compare to :py:meth:`__call__`, which applies the
        histogrammer to a single replica for efficiency.
        """
        replicated = sc.concat([var] * self.replicas, dim=self._replica_dim)
        return self._hist(replicated, coords=self._coords) / self.replicas


class Detector:
    def __init__(self, detector_number: sc.Variable):
        self._data = sc.DataArray(
            sc.zeros(sizes=detector_number.sizes, unit='counts', dtype='int32'),
            coords={'detector_number': detector_number},
        )
        self._detector_number = detector_number
        self._flat_detector_number = detector_number.flatten(to='event_id')
        self._start = int(self._flat_detector_number[0].value)
        self._stop = int(self._flat_detector_number[-1].value)
        self._size = int(self._flat_detector_number.size)
        self._sorted = sc.issorted(self._flat_detector_number, dim='event_id')
        self._consecutive = self._stop - self._start + 1 == self._size

    @property
    def detector_number(self) -> sc.Variable:
        return self._detector_number

    @property
    def data(self) -> sc.DataArray:
        return self._data

    def bincount(self, data: Sequence[int]) -> sc.DataArray:
        if not self._sorted:
            raise ValueError("Detector numbers must be sorted to use `bincount`.")
        if not self._consecutive:
            raise ValueError("Detector numbers must be consecutive to use `bincount`.")
        offset = np.asarray(data, dtype=np.int32) - self._start
        # Ignore events with detector numbers outside the range of the detector. This
        # should not happen in valid files but for now it is useful until we are sure
        # we get only valid files from upstream.
        offset = offset[(offset >= 0) & (offset < self._size)]
        out = sc.empty_like(self.data)
        out.values = np.bincount(offset, minlength=self._size).reshape(self.data.shape)
        return out

    def add_counts(self, data: Sequence[int]) -> None:
        self._data += self.bincount(data)

    def clear_counts(self) -> None:
        self._data.values *= 0


class RollingDetectorView(Detector):
    """
    Rolling view of a detector.

    The view keeps a history of counts and allows to access the sum of counts over a
    window of the most recent counts.
    """

    def __init__(
        self,
        *,
        detector_number: sc.Variable,
        window: int,
        projection: Callable[[sc.DataArray], sc.DataArray] | None = None,
    ):
        """
        Create a rolling detector view.

        Parameters
        ----------
        detector_number:
            Detector number for each pixel. This is used for mapping event_id to pixel.
        window:
            Size of the rolling window. This is the maximum window size, smaller windows
            can be accessed using the `get` method.
        projection:
            Optional projection function to apply to the counts before storing them in
            the history. This can be used to project the data onto a different
            coordinate system or to reduce the dimensionality of the data.
        """
        super().__init__(detector_number=detector_number)
        self._projection = projection
        self._window = window
        self._current = 0
        self._history: sc.DataArray
        self._cache: sc.DataArray
        self._cumulative: sc.DataArray
        self.clear_counts()

    @property
    def max_window(self) -> int:
        return self._window

    @property
    def cumulative(self) -> sc.DataArray:
        return self._cumulative

    def clear_counts(self) -> None:
        """
        Clear counts.

        Overrides Detector.clear_counts, to properly clear sliding window history and
        cache.
        """
        counts = sc.zeros_like(self.data)
        if self._projection is not None:
            counts = self._projection(counts)
        self._history = (
            sc.zeros_like(counts)
            .broadcast(sizes={'window': self._window, **counts.sizes})
            .copy()
        )
        self._cache = self._history.sum('window')
        self._cumulative = sc.zeros_like(self._cache)

    def make_roi_filter(self) -> roi.ROIFilter:
        """Return a ROI filter operating via the projection plane of the view."""
        norm = 1.0
        if isinstance(self._projection, Histogrammer):
            indices = self._projection.input_indices()
            norm = self._projection.replicas
        else:
            indices = sc.ones(sizes=self.data.sizes, dtype='int32', unit=None)
            indices = sc.cumsum(indices, mode='exclusive')
            if self._projection is not None:
                indices = self._projection(indices)
        return roi.ROIFilter(indices=indices, norm=norm)

    def transform_weights(
        self,
        weights: sc.Variable | sc.DataArray | None = None,
        *,
        threshold: float = 0.1,
    ) -> sc.DataArray:
        """
        Transform raw pixel weights to the projection plane.

        Parameters
        ----------
        weights:
            Raw pixel weights to transform. If None, default weights of 1 are used.
        threshold:
            Threshold for identifying bins with a low weight. If the weight is below the
            threshold times the median weight, the bin is marked as invalid. This is
            relevant to avoid issues with color scales in plots, where noise in bins
            with low weight may dominate the color scale if auto-scaling is used.
        """
        if weights is None:
            weights = sc.ones(
                sizes=self.detector_number.sizes, dtype='float32', unit=''
            )
        else:
            if weights.sizes != self.detector_number.sizes:
                raise sc.DimensionError(
                    f'Invalid {weights.sizes=} for {self.detector_number.sizes=}.'
                )
            if isinstance(weights, sc.DataArray):
                if (det_num := weights.coords.get('detector_number')) is not None:
                    if not sc.identical(det_num, self.detector_number):
                        raise sc.CoordError("Mismatching detector numbers in weights.")
                weights = weights.data
        if isinstance(self._projection, Histogrammer):
            xs = self._projection.apply_full(weights)  # Use all replicas
        elif self._projection is not None:
            xs = self._projection(weights)
        else:
            xs = weights.copy()
        nonempty = xs.values[xs.values > 0]
        mask = xs.values < threshold * np.median(nonempty)
        xs.values[mask] = np.nan
        return xs if isinstance(xs, sc.DataArray) else sc.DataArray(xs)

    @staticmethod
    def from_detector_and_histogrammer(
        detector: CalibratedDetector[SampleRun],
        window: RollingDetectorViewWindow,
        projection: Histogrammer,
    ) -> RollingDetectorView:
        """Helper for constructing via a Sciline workflow."""
        return RollingDetectorView(
            detector_number=detector.coords['detector_number'],
            window=window,
            projection=projection,
        )

    @staticmethod
    def from_detector_with_projection(
        projection: Callable[[sc.DataArray], sc.DataArray] | None,
    ) -> Callable[
        [CalibratedDetector[SampleRun], RollingDetectorViewWindow], RollingDetectorView
    ]:
        def factory(
            detector: CalibratedDetector[SampleRun],
            window: RollingDetectorViewWindow,
        ) -> RollingDetectorView:
            """Helper for constructing via a Sciline workflow."""
            return RollingDetectorView(
                detector_number=detector.coords['detector_number'],
                window=window,
                projection=projection,
            )

        return factory

    @staticmethod
    def from_nexus(
        nexus_file: str,
        *,
        detector_name: str,
        window: int,
        projection: Literal['xy_plane', 'cylinder_mantle_z']
        | Callable[[sc.DataArray], sc.DataArray]
        | None = None,
        resolution: dict[str, int] | None = None,
        pixel_noise: Literal['cylindrical'] | sc.Variable | None = None,
    ) -> RollingDetectorView:
        """
        Create a rolling detector view from a NeXus file using GenericNeXusWorkflow.

        The configuration parameters are preliminary for testing and are expected to
        change in the future.

        Parameters
        ----------
        nexus_file:
            NeXus file providing detector parameters such as detector_number and pixel
            positions.
        detector_name:
            Name of the detector in the NeXus file.
        window:
            Size of the rolling window.
        projection:
            Projection to use for the detector data. This can be a string selecting a
            predefined projection or a function that takes a DataArray and returns a
            DataArray. The predefined projections are 'xy_plane' and
            'cylinder_mantle_z'.
        resolution:
            Resolution to use for histogramming the detector data. Only required for
            'xy_plane' and 'cylinder_mantle_z' projections.
        pixel_noise:
            Noise to add to the pixel positions. This can be a scalar value to add
            Gaussian noise to the pixel positions or the string 'cylindrical' to add
            noise to the pixel positions of a cylindrical detector. Adding noise can be
            useful to avoid artifacts when projecting the data.
        """
        if pixel_noise is None:
            pixel_noise = sc.scalar(0.0, unit='m')
            noise_replica_count = 0
        else:
            # Unclear what a good number is, could be made configurable.
            noise_replica_count = 4
        wf = GenericNeXusWorkflow(run_types=[SampleRun], monitor_types=[])
        wf[RollingDetectorViewWindow] = window
        if projection == 'cylinder_mantle_z':
            wf.insert(make_cylinder_mantle_coords)
            wf.insert(RollingDetectorView.from_detector_and_histogrammer)
            wf[DetectorViewResolution] = resolution
        elif projection == 'xy_plane':
            wf.insert(make_xy_plane_coords)
            wf.insert(RollingDetectorView.from_detector_and_histogrammer)
            wf[DetectorViewResolution] = resolution
        else:
            wf[NeXusTransformation[snx.NXdetector, SampleRun]] = NeXusTransformation[
                snx.NXdetector, SampleRun
            ](sc.scalar(1))
            wf.insert(
                RollingDetectorView.from_detector_with_projection(projection=projection)
            )
        if isinstance(pixel_noise, sc.Variable):
            wf.insert(gaussian_position_noise)
            wf[PositionNoiseSigma] = pixel_noise
        elif pixel_noise == 'cylindrical':
            wf.insert(pixel_shape)
            wf.insert(pixel_cylinder_axis)
            wf.insert(pixel_cylinder_radius)
            wf.insert(position_noise_for_cylindrical_pixel)
        else:
            raise ValueError(f"Invalid {pixel_noise=}.")
        wf.insert(position_with_noisy_replicas)
        wf.insert(Histogrammer.from_coords)
        wf[PositionNoiseReplicaCount] = noise_replica_count
        wf[Filename[SampleRun]] = nexus_file
        wf[NeXusDetectorName] = detector_name
        return wf.compute(RollingDetectorView)

    def get(self, *, window: int | None = None) -> sc.DataArray:
        """
        Get the sum of counts over a window of the most recent counts.

        Parameters
        ----------
        window:
            Size of the window to use. If None, the full history is used.

        Returns
        -------
        :
            Sum of counts over the window.
        """
        if window is not None and not 0 <= window <= self._window:
            raise ValueError("Window size must be less than the history size.")
        if window is None or window == self._window:
            data = self._cache
        else:
            start = self._current - window
            if start >= 0:
                data = self._history['window', start : self._current].sum('window')
            else:
                data = self._history['window', start % self._window :].sum('window')
                data += self._history['window', 0 : self._current].sum('window')
        return data

    def add_events(self, data: sc.DataArray) -> None:
        """
        Add counts in the form of events grouped by pixel ID.

        Parameters
        ----------
        data:
            Events grouped by pixel ID, given by binned data.
        """
        counts = data.bins.size().to(dtype='int32', copy=False)
        counts.unit = 'counts'
        self._add_counts(counts)

    def add_counts(self, data: Sequence[int]) -> None:
        """
        Add counts in the form of a sequence of pixel IDs.

        Parameters
        ----------
        data:
            List of pixel IDs.
        """
        counts = self.bincount(data)
        self._add_counts(counts)

    def _add_counts(self, counts: sc.Variable) -> None:
        if self._projection is not None:
            counts = self._projection(counts)
        self._cache -= self._history['window', self._current]
        self._history['window', self._current] = counts
        self._cache += counts
        self._cumulative += counts
        self._current = (self._current + 1) % self._window


def project_xy(
    position: sc.Variable, *, zplane: sc.Variable | None = None
) -> dict[str, sc.Variable]:
    """
    Project positions onto the x-y plane, i.e., perpendicular to the beam (z axis).

    This is useful, e.g., for SANS instruments.
    """
    z = position.fields.z
    zmin = z.min()
    zero = sc.zeros_like(zmin)
    if zplane is None:
        zplane = z.max() if zmin < zero else zmin
    t = zplane / z
    return sc.DataGroup(x=position.fields.x * t, y=position.fields.y * t, z=zplane)


def project_onto_cylinder_z(
    position: sc.Variable, *, radius: sc.Variable | None = None
) -> dict[str, sc.Variable]:
    """
    Project positions onto the mantle of a cylinder aligned with the z axis.

    This is useful for cylindrical detectors, provided they are aligned along the beam.
    """
    x = position.fields.x
    y = position.fields.y
    r_xy = sc.sqrt(x**2 + y**2)
    if radius is None:
        radius = r_xy.min()
    t = radius / r_xy
    phi = sc.atan2(y=y, x=x).to(unit='deg')
    arc_length = radius * (phi * sc.scalar(np.pi / 180.0, unit='1/deg'))
    return sc.DataGroup(
        phi=phi, r=radius, z=position.fields.z * t, arc_length=arc_length
    )


def pixel_shape(component: NeXusComponent[snx.NXdetector, SampleRun]) -> PixelShape:
    return PixelShape(component['pixel_shape'])


def pixel_cylinder_axis(
    shape: PixelShape,
    transform: NeXusTransformation[snx.NXdetector, SampleRun],
) -> PixelCylinderAxis:
    vertices = shape['vertices']
    if len(vertices) != 3:
        raise NotImplementedError("Case of multiple cylinders not implemented.")
    # Note that transformation may be affine, so we need to apply it to the vertices
    # *before* subtracting them, to remove the translation part.
    t = transform.value
    return PixelCylinderAxis(t * vertices[2] - t * vertices[0])


def pixel_cylinder_radius(
    shape: PixelShape,
    transform: NeXusTransformation[snx.NXdetector, SampleRun],
) -> PixelCylinderRadius:
    vertices = shape['vertices']
    if len(vertices) != 3:
        raise NotImplementedError("Case of multiple cylinders not implemented.")
    # Note that transformation may be affine, so we need to apply it to the vertices
    # *before* subtracting them, to remove the translation part.
    t = transform.value
    return PixelCylinderRadius(t * vertices[1] - t * vertices[0])


# Arbitrary small but not tiny size. More is likely not necessary, but having it too
# small might lead to visible patterns after projection. It is therefore also chosen
# to be a prime number, but it likely does not matter.
_noise_size = 107


def position_noise_for_cylindrical_pixel(
    *, axis: PixelCylinderAxis, radius: PixelCylinderRadius
) -> PositionNoise:
    # We *assume* that the cylinder is centered on the origin. Real files may not
    # fulfill this. However, the rest of the data reduction currently assumes that
    # the pixel offset corresponds to the pixel center, so if it is not fulfilled
    # there are bigger problems elsewhere anywhere.
    rng = np.random.default_rng()
    dims = ('position',)
    size = _noise_size
    z_hat = axis / sc.norm(axis)  # Unit vector along the cylinder axis
    x_hat = radius / sc.norm(radius)  # Unit vector along the radius direction
    y_hat = sc.cross(z_hat, x_hat)  # Unit vector perpendicular to z_hat and x_hat

    dz = sc.array(dims=dims, values=rng.uniform(-0.5, 0.5, size=size)) * axis
    dphi = sc.array(dims=dims, values=rng.uniform(0, 2 * np.pi, size=size), unit='rad')
    r = sc.norm(radius).value
    dr = sc.sqrt(
        sc.array(dims=dims, values=rng.uniform(0, r**2, size=size), unit='m^2')
    )

    dx = dr * sc.cos(dphi) * x_hat
    dy = dr * sc.sin(dphi) * y_hat

    return PositionNoise(dx + dy + dz)


def gaussian_position_noise(sigma: PositionNoiseSigma) -> PositionNoise:
    sigma = sigma.to(unit='m', copy=False)
    size = _noise_size
    position = sc.empty(sizes={'position': size}, unit='m', dtype=sc.DType.vector3)
    position.values = np.random.default_rng(seed=1234).normal(
        0, sigma.value, size=(size, 3)
    )
    return PositionNoise(position)


def position_with_noisy_replicas(
    *,
    detector: CalibratedDetector[SampleRun],
    position_noise: PositionNoise,
    replicas: PositionNoiseReplicaCount,
) -> CalibratedPositionWithNoisyReplicas:
    """
    Create a new position array with noise added to the detector positions.

    The first slice of the new array is the original position data, and the
    remaining slices are the original data with noise added.
    """
    position = detector.coords['position'].to(unit='m')
    if replicas == 0:
        return sc.concat([position], dim='replica')
    noise_dim = position_noise.dim
    size = position.size * replicas
    # "Paint" the short array of noise on top of the (replicated) position data.
    noise = (
        sc.broadcast(
            position_noise,
            sizes={'dummy': ceil(size / position_noise.size), **position_noise.sizes},
        )
        .flatten(to=noise_dim)[:size]
        .fold(dim=noise_dim, sizes={'replica': replicas, **position.sizes})
    )
    return sc.concat([position, noise + position], dim='replica')


def make_xy_plane_coords(
    position: CalibratedPositionWithNoisyReplicas,
) -> ProjectedCoords:
    # The first slice is the original data, so we use it to determine the z plane.
    # This avoids noise in the z plane which could later cause trouble when
    # combining the data.
    zplane = project_xy(position['replica', 0])['z']
    return project_xy(position, zplane=zplane)


def make_cylinder_mantle_coords(
    position: CalibratedPositionWithNoisyReplicas,
) -> ProjectedCoords:
    radius = project_onto_cylinder_z(position['replica', 0])['r']
    return project_onto_cylinder_z(position, radius=radius)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401
from functools import singledispatch
from typing import Any, Protocol

import ipywidgets as widgets

from ..parameter import (
    BinEdgesParameter,
    BooleanParameter,
    FilenameParameter,
    MultiFilenameParameter,
    MultiStringParameter,
    Parameter,
    ParamWithBounds,
    ParamWithOptions,
    StringParameter,
    Vector2dParameter,
    Vector3dParameter,
)
from ._binedges_widget import BinEdgesWidget
from ._bounds_widget import BoundsWidget
from ._config import default_layout, default_style
from ._filename_widget import FilenameWidget, MultiFilenameWidget
from ._linspace_widget import LinspaceWidget
from ._optional_widget import OptionalWidget
from ._spinner import Spinner
from ._string_widget import MultiStringWidget, StringWidget
from ._switchable_widget import SwitchWidget
from ._vector_widget import VectorWidget


class EssWidget(Protocol):
    """Protocol for ESS widgets.

    All widgets should have a `value` property that returns the value of the widget.
    It can be composed from multiple widgets.
    """

    @property
    def value(self) -> Any: ...


from collections.abc import Callable
from functools import wraps


def switchable_widget(
    func: Callable[[Parameter], widgets.Widget],
) -> Callable[[Parameter], widgets.Widget]:
    """Wrap a widget in a switchable widget."""

    @wraps(func)
    def wrapper(param: Parameter) -> widgets.Widget:
        widget = func(param)
        if param.switchable:
            return SwitchWidget(widget, name=param.name)
        return widget

    return wrapper


def optional_widget(
    func: Callable[[Parameter], widgets.Widget],
) -> Callable[[Parameter], widgets.Widget]:
    """Wrap a widget in a optional widget."""

    @wraps(func)
    def wrapper(param: Parameter) -> widgets.Widget:
        widget = func(param)
        if param.optional:
            return OptionalWidget(widget, name=param.name)
        return widget

    return wrapper


@switchable_widget
@optional_widget  # optional_widget should be applied first
@singledispatch
def create_parameter_widget(param: Parameter) -> widgets.Widget:
    """Create a widget for a parameter depending on the ``param`` type.

    If the type of the parameter is not supported, a text widget is returned.
    """
    return widgets.Text(
        '', description=param.name, layout=default_layout, style=default_style
    )


@create_parameter_widget.register(BooleanParameter)
def boolean_parameter_widget(param: BooleanParameter):
    name = param.name.split('.')[-1]
    description = param.description
    return widgets.Checkbox(
        value=param.default,
        description=name,
        tooltip=description,
        layout=default_layout,
        style=default_style,
    )


@create_parameter_widget.register(StringParameter)
def string_parameter_widget(param: StringParameter):
    return StringWidget(
        description=param.name,
        value=param.default,
        layout=default_layout,
        style=default_style,
    )


@create_parameter_widget.register(MultiStringParameter)
def multi_string_parameter_widget(param: MultiStringParameter):
    return MultiStringWidget(
        description=param.name,
        value=param.default,
        layout=default_layout,
        style=default_style,
    )


@create_parameter_widget.register(FilenameParameter)
def filename_parameter_widget(param: FilenameParameter):
    return FilenameWidget(
        description=param.name,
        value=param.default,
        layout=default_layout,
        style=default_style,
    )


@create_parameter_widget.register(MultiFilenameParameter)
def multi_filename_parameter_widget(param: MultiFilenameParameter):
    return MultiFilenameWidget(
        description=param.name,
        value=param.default,
        layout=default_layout,
        style=default_style,
    )


@create_parameter_widget.register(ParamWithOptions)
def param_with_option_widget(param: ParamWithOptions):
    return widgets.Dropdown(
        description=param.name,
        options=param.options,
        layout=default_layout,
        style=default_style,
    )


@create_parameter_widget.register(ParamWithBounds)
def param_with_bounds_widget(param: ParamWithBounds):
    return BoundsWidget()


@create_parameter_widget.register(BinEdgesParameter)
def bin_edges_parameter_widget(param: BinEdgesParameter):
    return BinEdgesWidget(
        name=param.name,
        dim=param.dim,
        start=param.start,
        stop=param.stop,
        nbins=param.nbins,
        unit=param.unit,
        log=param.log,
    )


@create_parameter_widget.register(Vector2dParameter)
def vector_2d_parameter_widget(param: Vector2dParameter):
    return VectorWidget(name=param.name, variable=param.default, components="xy")


@create_parameter_widget.register(Vector3dParameter)
def vector_3d_parameter_widget(param: Vector3dParameter):
    return VectorWidget(name=param.name, variable=param.default, components="xyz")


__all__ = [
    'BinEdgesWidget',
    'BoundsWidget',
    'EssWidget',
    'FilenameWidget',
    'LinspaceWidget',
    'MultiFilenameWidget',
    'OptionalWidget',
    'Spinner',
    'SwitchWidget',
    'VectorWidget',
    'create_parameter_widget',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import warnings
from collections.abc import Iterable
from typing import Any, Protocol, runtime_checkable

from ipywidgets import Widget


@runtime_checkable
class WidgetWithFieldsProtocol(Protocol):
    def set_fields(self, new_values: dict[str, Any]) -> None: ...

    def get_fields(self) -> dict[str, Any]: ...


def _warn_invalid_field(invalid_fields: Iterable[str]) -> None:
    for field_name in invalid_fields:
        warning_msg = f"Cannot set field '{field_name}'."
        " The field does not exist in the widget."
        "The field value will be ignored."
        warnings.warn(warning_msg, UserWarning, stacklevel=2)


class WidgetWithFieldsMixin:
    def set_fields(self, new_values: dict[str, Any]) -> None:
        # Extract valid fields
        new_field_names = set(new_values.keys())
        valid_field_names = new_field_names & set(self.fields.keys())
        # Warn for invalid fields
        invalid_field_names = new_field_names - valid_field_names
        _warn_invalid_field(invalid_field_names)
        # Set valid fields
        for field_name in valid_field_names:
            self.fields[field_name].value = new_values[field_name]

    def get_fields(self) -> dict[str, Any]:
        return {
            field_name: field_sub_widget.value
            for field_name, field_sub_widget in self.fields.items()
        }


def set_fields(widget: Widget, new_values: dict[str, Any]) -> None:
    """Set the fields of a widget with the given values.

    Parameters
    ----------
    widget:
        The widget to set the fields. It should either be an instance of
        ``WidgetWithFieldsProtocol`` or have a value property setter.
    new_values:
        The new values to set for the fields.
        i.e. ``{'field_name': field_value}``
        If the widget does not have a ``set_fields/get_fields`` method,
        (e.g. it is not an instance of ``WidgetWithFieldsProtocol``),
        it will try to set the value of the widget directly.
        The value of the widget should be available in the ``new_values`` dictionary
        with the key 'value'.
        i.e. ``{'value': widget_value}``

    Raises
    ------
    TypeError:
        If ``new_values`` is not a dictionary.

    """
    if not isinstance(new_values, dict):
        raise TypeError(f"new_values must be a dictionary, got {type(new_values)}")

    if isinstance(widget, WidgetWithFieldsProtocol) and isinstance(new_values, dict):
        widget.set_fields(new_values)
    else:
        try:
            # Use value property setter if ``new_values`` contains 'value'
            if 'value' in new_values:
                widget.value = new_values['value']
            # Warn if there is any other fields in new_values
            _warn_invalid_field(set(new_values.keys()) - {'value'})
        except AttributeError as error:
            # Checking if the widget value property has a setter in advance, i.e.
            # ```python
            # (widget_property := getattr(type(widget), 'value', None)) is not None
            # and getattr(widget_property, 'fset', None) is not None
            # ```
            # does not work with a class that inherits Traitlets class.
            # In those classes, even if a property has a setter,
            # it may not have `fset` attribute.
            # It is not really feasible to check all possible cases of value setters.
            # Instead, we try setting the value and catch the AttributeError.
            # to determine if the widget has a value setter.
            warnings.warn(
                f"Cannot set value for widget of type {type(widget)}."
                " The new_value(s) will be ignored."
                f" Setting value caused the following error: {error}",
                UserWarning,
                stacklevel=1,
            )


def get_fields(widget: Widget) -> dict[str, Any] | None:
    """Get the fields of a widget.

    If the widget is an instance of ``WidgetWithFieldsProtocol``,
    it will return the fields of the widget.
    i.e. ``{'field_name': field_value}``
    Otherwise, it will try to get the value of the widget and return a dictionary
    with the key 'value' and the value of the widget.
    i.e. ``{'value': widget_value}``

    Parameters
    ----------
    widget:
        The widget to get the fields. It should either be an instance of
        ``WidgetWithFieldsProtocol`` or have a value property.

    """
    if isinstance(widget, WidgetWithFieldsProtocol):
        return widget.get_fields()
    try:
        return {'value': widget.value}
    except AttributeError:
        warnings.warn(
            f"Cannot get value or fields for widget of type {type(widget)}.",
            UserWarning,
            stacklevel=1,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_identical

from ess.reduce.time_of_flight import resample


class TestFindStrictlyIncreasingSections:
    """Tests for find_strictly_increasing_sections function."""

    def test_given_monotonically_increasing_values_returns_single_section(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 4, 5])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(0, 5)]

    def test_given_multiple_increasing_sections_returns_corresponding_slices(self):
        var = sc.array(dims=['x'], values=[1, 3, 2, 4, 6, 5, 7, 9])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(0, 2), slice(2, 5), slice(5, 8)]

    def test_given_flat_sections_finds_strictly_increasing_parts_only(self):
        var = sc.array(dims=['x'], values=[1, 2, 2, 3, 4, 4, 5])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(0, 2), slice(2, 5), slice(5, 7)]

    def test_given_extended_flat_sections_finds_strictly_increasing_parts_only(self):
        var = sc.array(dims=['x'], values=[1, 2, 2, 2, 3, 4, 4, 5])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(0, 2), slice(3, 6), slice(6, 8)]

    def test_given_decreasing_values_returns_empty_list(self):
        var = sc.array(dims=['x'], values=[5, 4, 3, 2, 1])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == []

    def test_given_mixed_behavior_finds_only_increasing_sections(self):
        var = sc.array(dims=['x'], values=[1, 3, 5, 4, 3, 6, 7, 8])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(0, 3), slice(4, 8)]

    def test_given_empty_array_returns_empty_list(self):
        var = sc.array(dims=['x'], values=[])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == []

    def test_given_single_value_returns_empty_list(self):
        var = sc.array(dims=['x'], values=[42])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == []

    def test_given_two_increasing_values_returns_single_section(self):
        var = sc.array(dims=['x'], values=[1, 2])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(0, 2)]

    def test_given_two_equal_values_returns_empty_list(self):
        var = sc.array(dims=['x'], values=[2, 2])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == []

    def test_given_nan_in_middle_stops_section_at_nan(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, np.nan, 5, 6])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(0, 3), slice(4, 6)]

    def test_given_nan_at_beginning_skips_nan(self):
        var = sc.array(dims=['x'], values=[np.nan, 2, 3, 4])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(1, 4)]

    def test_given_nan_at_end_stops_section_before_nan(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, np.nan])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(0, 3)]

    def test_given_multiple_nans_finds_valid_sections_between(self):
        var = sc.array(dims=['x'], values=[1, np.nan, 3, 4, np.nan, 6, 7])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(2, 4), slice(5, 7)]

    def test_given_all_nans_returns_empty_list(self):
        var = sc.array(dims=['x'], values=[np.nan, np.nan, np.nan])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == []

    def test_given_increasing_section_at_beginning_returns_correct_slice(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 3, 2, 1])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(0, 3)]

    def test_given_increasing_section_in_middle_returns_correct_slice(self):
        var = sc.array(dims=['x'], values=[3, 2, 1, 2, 3, 4, 3, 2])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(2, 6)]

    def test_given_increasing_section_at_end_returns_correct_slice(self):
        var = sc.array(dims=['x'], values=[3, 3, 2, 3, 4, 5])
        sections = resample.find_strictly_increasing_sections(var)
        assert sections == [slice(2, 6)]


class TestGetMinMax:
    """Tests for get_min_max function."""

    def test_basic_functionality(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 2, 3, 4, 5, 3, 4, 5, 6])
        slices = [slice(0, 3), slice(3, 7), slice(7, 11)]
        min_val, max_val = resample.get_min_max(var, dim='x', slices=slices)
        assert min_val.value == 1
        assert max_val.value == 6

    def test_with_units(self):
        var = sc.array(dims=['x'], values=[1.0, 2.0, 3.0, 2.0, 3.0, 4.0], unit='m')
        slices = [slice(0, 3), slice(3, 6)]
        min_val, max_val = resample.get_min_max(var, dim='x', slices=slices)
        assert min_val.value == 1.0
        assert max_val.value == 4.0
        assert min_val.unit == sc.Unit('m')
        assert max_val.unit == sc.Unit('m')

    def test_with_single_slice(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 4, 5])
        slices = [slice(0, 5)]
        min_val, max_val = resample.get_min_max(var, dim='x', slices=slices)
        assert min_val.value == 1
        assert max_val.value == 5

    def test_with_non_contiguous_slices(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 4, 5, 0, 1, 2])
        slices = [slice(0, 5), slice(5, 8)]
        min_val, max_val = resample.get_min_max(var, dim='x', slices=slices)
        assert min_val.value == 0  # The min value is at the start of the second slice
        assert max_val.value == 5  # The max value is at the end of the first slice

    def test_with_overlapping_slices(self):
        var = sc.array(dims=['x'], values=[1, 3, 5, 7, 9])
        slices = [slice(0, 3), slice(1, 4), slice(2, 5)]
        min_val, max_val = resample.get_min_max(var, dim='x', slices=slices)
        assert min_val.value == 1  # The min value is at the start of the first slice
        assert max_val.value == 9  # The max value is at the end of the third slice

    def test_with_different_dimension_name(self):
        var = sc.array(dims=['time'], values=[10, 20, 30, 40, 50])
        slices = [slice(1, 3), slice(3, 5)]
        min_val, max_val = resample.get_min_max(var, dim='time', slices=slices)
        assert min_val.value == 20  # The min value is at the start of the first slice
        assert max_val.value == 50  # The max value is at the end of the second slice

    def test_with_float_data(self):
        var = sc.array(dims=['x'], values=[1.1, 2.2, 3.3, 4.4, 5.5])
        slices = [slice(0, 3), slice(3, 5)]
        min_val, max_val = resample.get_min_max(var, dim='x', slices=slices)
        assert min_val.value == 1.1
        assert max_val.value == 5.5

    def test_with_mixed_positive_and_negative_values(self):
        var = sc.array(dims=['x'], values=[-5, -3, -1, 0, 2, 4])
        slices = [slice(0, 3), slice(3, 6)]
        min_val, max_val = resample.get_min_max(var, dim='x', slices=slices)
        assert min_val.value == -5
        assert max_val.value == 4

    def test_with_empty_slices_list_raises_error(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 4, 5])
        slices = []
        with pytest.raises(ValueError, match="No strictly increasing sections found."):
            resample.get_min_max(var, dim='x', slices=slices)

    def test_integration_with_find_strictly_increasing_sections(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 2, 3, 4, 1, 2, 3])
        slices = resample.find_strictly_increasing_sections(var)
        min_val, max_val = resample.get_min_max(var, dim='x', slices=slices)
        assert min_val.value == 1
        assert max_val.value == 4


class TestMakeRegularGrid:
    """Tests for make_regular_grid function."""

    def test_basic_functionality(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 2, 3, 4])
        slices = [slice(0, 3), slice(3, 6)]
        grid = resample.make_regular_grid(var, dim='x', slices=slices)
        assert sc.identical(grid, sc.array(dims=['x'], values=[1, 2, 3, 4]))

    def test_with_units(self):
        var = sc.array(dims=['x'], values=[1.0, 2.0, 3.0, 1.5, 2.5, 3.5], unit='m')
        slices = [slice(0, 3), slice(3, 6)]
        grid = resample.make_regular_grid(var, dim='x', slices=slices)
        expected = sc.array(dims=['x'], values=[1.0, 2.0, 3.0, 4.0], unit='m')
        assert sc.identical(grid, expected)
        assert grid.unit == sc.Unit('m')

    def test_with_different_step_sizes_uses_minimum_section(self):
        # First section has step size 1, second has step size 0.5
        var = sc.array(dims=['x'], values=[1, 2, 3, 3.5, 4, 4.5])
        slices = [slice(0, 3), slice(3, 6)]
        grid = resample.make_regular_grid(var, dim='x', slices=slices)
        # Should use step size 1 from the first section (where min value is located)
        expected = sc.array(dims=['x'], values=[1.0, 2, 3, 4, 5])
        assert sc.identical(grid, expected)

    def test_with_different_dimension_name(self):
        var = sc.array(dims=['time'], values=[10, 20, 30, 5, 15, 25])
        slices = [slice(0, 3), slice(3, 6)]
        grid = resample.make_regular_grid(var, dim='time', slices=slices)
        expected = sc.array(dims=['time'], values=[5, 15, 25, 35])
        assert sc.identical(grid, expected)

    def test_with_float_data(self):
        var = sc.array(dims=['x'], values=[1.1, 2.2, 3.3, 0.5, 1.6, 2.7])
        slices = [slice(0, 3), slice(3, 6)]
        grid = resample.make_regular_grid(var, dim='x', slices=slices)
        expected = sc.array(dims=['x'], values=[0.5, 1.6, 2.7, 3.8])
        assert sc.allclose(grid, expected)

    def test_with_mixed_positive_and_negative_values(self):
        var = sc.array(dims=['x'], values=[-5, -3, -1, 0, 2, 4])
        slices = [slice(0, 3), slice(3, 6)]
        grid = resample.make_regular_grid(var, dim='x', slices=slices)
        expected = sc.array(dims=['x'], values=[-5, -3, -1, 1, 3, 5])
        assert sc.identical(grid, expected)

    def test_with_single_slice(self):
        var = sc.array(dims=['x'], values=[1, 3, 5, 7, 9])
        slices = [slice(0, 5)]
        grid = resample.make_regular_grid(var, dim='x', slices=slices)
        expected = sc.array(dims=['x'], values=[1, 3, 5, 7, 9])
        assert sc.identical(grid, expected)

    def test_with_different_dtype(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 2, 3, 4], dtype=np.int64)
        slices = [slice(0, 3), slice(3, 6)]
        grid = resample.make_regular_grid(var, dim='x', slices=slices)
        expected = sc.array(dims=['x'], values=[1, 2, 3, 4], dtype=np.int64)
        assert sc.identical(grid, expected)
        assert grid.dtype == np.int64

    def test_empty_slices_list_raises_error(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 4, 5])
        slices = []
        with pytest.raises(ValueError, match="No strictly increasing sections found."):
            resample.make_regular_grid(var, dim='x', slices=slices)

    def test_integration_with_find_strictly_increasing_sections(self):
        var = sc.array(dims=['x'], values=[1, 3, 5, 2, 4, 6])
        slices = resample.find_strictly_increasing_sections(var)
        grid = resample.make_regular_grid(var, dim='x', slices=slices)
        expected = sc.array(dims=['x'], values=[1, 3, 5, 7])
        assert sc.identical(grid, expected)

    def test_ensures_last_bin_edge_is_included(self):
        var = sc.array(dims=['x'], values=[1, 2, 3, 2, 3, 4, 5])
        slices = [slice(0, 3), slice(3, 7)]
        grid = resample.make_regular_grid(var, dim='x', slices=slices)
        expected = sc.array(dims=['x'], values=[1, 2, 3, 4, 5])
        assert sc.identical(grid, expected)
        # Test that the maximum value from original data is included in the grid
        assert grid[-1].value == 5


class TestRebinStrictlyIncreasing:
    """Tests for rebin_strictly_increasing function."""

    def test_basic_functionality(self):
        # Create a data array with a simple time-of-flight coordinate that has two
        # strictly increasing sections
        tof = sc.array(dims=['tof'], values=[1, 2, 3, 2, 3, 4, 5])
        # 666 is in the "unphysical" negative-size bin and should be dropped.
        data = sc.array(dims=['tof'], values=[10, 20, 666, 25, 35, 11])
        da = sc.DataArray(data=data, coords={'tof': tof})

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Check the rebinned result has a regular grid from 1 to 5
        expected_tof = sc.array(dims=['tof'], values=[1.0, 2, 3, 4, 5])
        assert sc.identical(result.coords['tof'], expected_tof)

        # Check the data values are properly rebinned and combined
        expected_data = sc.array(dims=['tof'], values=[10.0, 20 + 25, 35, 11])
        assert_identical(result.data, expected_data)

    def test_with_different_step_sizes(self):
        # First section has step size 1, second has step size 0.5
        tof = sc.array(dims=['tof'], values=[1, 2, 4, 3.5, 4, 4.5, 5])
        data = sc.array(dims=['tof'], values=[10, 20, 15, 25, 35, 45])
        da = sc.DataArray(data=data, coords={'tof': tof})

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Should use step size 1 from the first section (where min value is found)
        expected_tof = sc.array(dims=['tof'], values=[1.0, 2, 3, 4, 5])
        assert_identical(result.coords['tof'], expected_tof)

    def test_with_units(self):
        tof = sc.array(dims=['tof'], values=[1.0, 2.0, 3.0, 2.0, 3.0, 4.0], unit='ms')
        data = sc.array(dims=['tof'], values=[10, 20, 15, 25, 35], unit='counts')
        da = sc.DataArray(data=data, coords={'tof': tof})

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Check units are preserved
        assert result.coords['tof'].unit == sc.Unit('ms')
        assert result.data.unit == sc.Unit('counts')

        # Check values
        expected_tof = sc.array(dims=['tof'], values=[1.0, 2.0, 3.0, 4.0], unit='ms')
        assert sc.identical(result.coords['tof'], expected_tof)

    def test_with_single_increasing_section(self):
        tof = sc.array(dims=['tof'], values=[1, 2, 3, 4, 5, 6])
        data = sc.array(dims=['tof'], values=[10, 20, 30, 40, 50])
        da = sc.DataArray(data=data, coords={'tof': tof})

        result = resample.rebin_strictly_increasing(da, 'tof')

        # For a single increasing section, should return just that section
        assert sc.identical(result, da)

    def test_with_single_section_minimum_length(self):
        """Test with a single section of the minimum length (2 points)."""
        tof = sc.array(dims=['tof'], values=[1, 2])
        data = sc.array(dims=['tof'], values=[10])
        da = sc.DataArray(data=data, coords={'tof': tof})

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Should return the original section without rebinning
        assert sc.identical(result, da)

    def test_with_single_section_integer_coords(self):
        """Test with a single section with integer coordinates."""
        tof = sc.array(dims=['tof'], values=[1, 2, 3, 4], dtype=np.int32)
        data = sc.array(dims=['tof'], values=[10, 20, 30])
        da = sc.DataArray(data=data, coords={'tof': tof})

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Should return the original section, with converted coords
        assert sc.identical(result, da)
        # Check that we're maintaining the same values even if the dtype changed
        assert sc.allclose(result.coords['tof'], da.coords['tof'])

    def test_with_single_section_with_masks_and_extra_coords(self):
        """Test with a single section that has masks and attributes."""
        tof = sc.array(dims=['tof'], values=[1, 2, 3, 4, 5])
        data = sc.array(dims=['tof'], values=[10, 20, 30, 40])
        da = sc.DataArray(data=data, coords={'tof': tof})

        # Add mask and attribute
        mask = sc.array(dims=['tof'], values=[False, True, False, False])
        da.masks['quality'] = mask
        da.coords['test_attr'] = sc.scalar(42)

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Should preserve the original data with masks and attributes
        assert sc.identical(result, da)
        assert 'quality' in result.masks
        assert sc.identical(result.masks['quality'], mask)
        assert 'test_attr' in result.coords
        assert sc.identical(result.coords['test_attr'], sc.scalar(42))

    def test_with_three_increasing_sections(self):
        tof = sc.array(dims=['tof'], values=[1, 2, 3, 1, 2, 3, 4, 2, 3, 4, 5])
        data = sc.array(dims=['tof'], values=[5, 10, 6, 12, 18, 8, 14, 21, 28, 35])
        da = sc.DataArray(data=data, coords={'tof': tof})

        result = resample.rebin_strictly_increasing(da, 'tof')

        expected_tof = sc.array(dims=['tof'], values=[1.0, 2, 3, 4, 5])
        assert_identical(result.coords['tof'], expected_tof)

        # Sum of all three sections properly rebinned
        expected_data = sc.array(
            dims=['tof'], values=[5.0 + 12, 10 + 18 + 21, 8 + 28, 35]
        )
        assert_identical(result.data, expected_data)

    def test_with_nan_values_in_coordinate(self):
        tof = sc.array(dims=['tof'], values=[1, 2, 3, np.nan, 5, 6, 7])
        data = sc.array(dims=['tof'], values=[10, 20, 40, 50, 60, 70])
        da = sc.DataArray(data=data, coords={'tof': tof})

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Should have two increasing sections: [1,2,3] and [5,6,7]
        expected_tof = sc.array(dims=['tof'], values=[1.0, 2, 3, 4, 5, 6, 7])
        assert_identical(result.coords['tof'], expected_tof)

        # Data should be correctly rebinned, excluding the NaN point
        expected_data = sc.array(dims=['tof'], values=[10.0, 20, 0, 0, 60, 70])
        assert_identical(result.data, expected_data)

    def test_with_no_increasing_sections_raises_error(self):
        tof = sc.array(dims=['tof'], values=[5, 4, 3, 2, 1, 0])
        data = sc.array(dims=['tof'], values=[10, 20, 30, 40, 50])
        da = sc.DataArray(data=data, coords={'tof': tof})

        with pytest.raises(ValueError, match="No strictly increasing sections found."):
            resample.rebin_strictly_increasing(da, 'tof')

    def test_with_variances(self):
        tof = sc.array(dims=['tof'], values=[1, 2, 3, 2, 3, 4])
        values = [
            10.0,
            20.0,
            15.0,
            25.0,
            35.0,
        ]  # Using float for values to match variances
        variances = [1.0, 2.0, 1.5, 2.5, 3.5]  # Float variances
        data = sc.array(dims=['tof'], values=values, variances=variances)
        da = sc.DataArray(data=data, coords={'tof': tof})

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Check that variances are properly propagated
        assert result.data.variances is not None
        expected_variances = sc.array(dims=['tof'], values=[1.0, 2.0 + 2.5, 3.5])
        assert_identical(sc.variances(result.data), expected_variances)

    def test_additional_coords_are_dropped(self):
        tof = sc.array(dims=['tof'], values=[1, 2, 3, 2, 3, 4, 5])
        data = sc.array(dims=['tof'], values=[10, 20, 15, 25, 35, 45])
        energy = sc.array(dims=['tof'], values=[1.1, 1.2, 1.3, 1.2, 1.4, 1.2])
        da = sc.DataArray(data=data, coords={'tof': tof, 'energy': energy})

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Rebin cannot preserve coords
        assert 'energy' not in result.coords

    def test_masks_are_applied(self):
        tof = sc.array(dims=['tof'], values=[1, 2, 3, 2, 3, 4, 5])
        data = sc.array(dims=['tof'], values=[10, 20, 15, 25, 35, 45])
        da = sc.DataArray(data=data, coords={'tof': tof})

        baseline = resample.rebin_strictly_increasing(da, 'tof')

        # Add a mask
        mask = sc.array(dims=['tof'], values=[False, False, True, False, False, True])
        da.masks['quality'] = mask

        result = resample.rebin_strictly_increasing(da, 'tof')
        # Rebin applies masks
        assert 'quality' not in baseline.masks
        assert result.sum().value < baseline.sum().value

    def test_uses_coord_name_as_new_dimension_name(self):
        """Test with a dimension name different from the coordinate name."""
        # Create data array with dimension name 'd' but coordinate named 'tof'
        tof = sc.array(dims=['d'], values=[1, 2, 3, 2, 3, 4])
        data = sc.array(dims=['d'], values=[10, 20, 15, 25, 35])
        da = sc.DataArray(data=data, coords={'tof': tof})

        # The dimension name 'd' is different from the coordinate name 'tof'
        assert da.coords['tof'].dims[0] == 'd'
        assert 'tof' not in da.dims

        result = resample.rebin_strictly_increasing(da, 'tof')

        # Check that the dimension has been renamed to match the coordinate
        assert 'tof' in result.dims
        assert result.coords['tof'].dims[0] == 'tof'

        # Check that rebinning worked correctly
        expected_tof = sc.array(dims=['tof'], values=[1.0, 2.0, 3.0, 4.0])
        assert sc.identical(result.coords['tof'], expected_tof)

        # Check the data values are properly rebinned and combined
        expected_data = sc.array(dims=['tof'], values=[10.0, 20.0 + 25.0, 35.0])
        assert_identical(result.data, expected_data)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""
Utilities for computing time-of-flight lookup tables from neutron simulations.
"""

import math
from dataclasses import dataclass
from typing import NewType

import numpy as np
import sciline as sl
import scipp as sc

from ..nexus.types import DiskChoppers
from .types import TimeOfFlightLookupTable


@dataclass
class SimulationResults:
    """
    Results of a time-of-flight simulation used to create a lookup table.

    The results (apart from ``distance``) should be flat lists (1d arrays) of length N
    where N is the number of neutrons, containing the properties of the neutrons in the
    simulation.

    Parameters
    ----------
    time_of_arrival:
        Time of arrival of the neutrons at the position where the events were recorded
        (1d array of size N).
    speed:
        Speed of the neutrons, typically derived from the wavelength of the neutrons
        (1d array of size N).
    wavelength:
        Wavelength of the neutrons (1d array of size N).
    weight:
        Weight/probability of the neutrons (1d array of size N).
    distance:
        Distance from the source to the position where the events were recorded
        (single value; we assume all neutrons were recorded at the same position).
        For a ``tof`` simulation, this is just the position of the detector where the
        events are recorded. For a ``McStas`` simulation, this is the distance between
        the source and the event monitor.
    """

    time_of_arrival: sc.Variable
    speed: sc.Variable
    wavelength: sc.Variable
    weight: sc.Variable
    distance: sc.Variable


NumberOfSimulatedNeutrons = NewType("NumberOfSimulatedNeutrons", int)
"""
Number of neutrons simulated in the simulation that is used to create the lookup table.
This is typically a large number, e.g., 1e6 or 1e7.
"""

LtotalRange = NewType("LtotalRange", tuple[sc.Variable, sc.Variable])
"""
Range (min, max) of the total length of the flight path from the source to the detector.
This is used to create the lookup table to compute the neutron time-of-flight.
Note that the resulting table will extend slightly beyond this range, as the supplied
range is not necessarily a multiple of the distance resolution.

Note also that the range of total flight paths is supplied manually to the workflow
instead of being read from the input data, as it allows us to compute the expensive part
of the workflow in advance (the lookup table) and does not need to be repeated for each
run, or for new data coming in in the case of live data collection.
"""

DistanceResolution = NewType("DistanceResolution", sc.Variable)
"""
Step size of the distance axis in the lookup table.
Should be a single scalar value with a unit of length.
This is typically of the order of 1-10 cm.
"""

TimeResolution = NewType("TimeResolution", sc.Variable)
"""
Step size of the event_time_offset axis in the lookup table.
This is basically the 'time-of-flight' resolution of the detector.
Should be a single scalar value with a unit of time.
This is typically of the order of 0.1-0.5 ms.

Since the event_time_offset range needs to span exactly one pulse period, the final
resolution in the lookup table will be at least the supplied value here, but may be
smaller if the pulse period is not an integer multiple of the time resolution.
"""


LookupTableRelativeErrorThreshold = NewType("LookupTableRelativeErrorThreshold", float)
"""
Threshold for the relative standard deviation (coefficient of variation) of the
projected time-of-flight above which values are masked.
"""

PulsePeriod = NewType("PulsePeriod", sc.Variable)
"""
Period of the source pulses, i.e., time between consecutive pulse starts.
"""

PulseStride = NewType("PulseStride", int)
"""
Stride of used pulses. Usually 1, but may be a small integer when pulse-skipping.
"""

SourcePosition = NewType("SourcePosition", sc.Variable)
"""
Position of the neutron source in the coordinate system of the choppers.
"""

SimulationSeed = NewType("SimulationSeed", int | None)
"""Seed for the random number generator used in the simulation.
"""

SimulationFacility = NewType("SimulationFacility", str)
"""
Facility where the experiment is performed, e.g., 'ess'.
"""


def _mask_large_uncertainty(table: sc.DataArray, error_threshold: float):
    """
    Mask regions with large uncertainty with NaNs.
    The values are modified in place in the input table.

    Parameters
    ----------
    table:
        Lookup table with time-of-flight as a function of distance and time-of-arrival.
    error_threshold:
        Threshold for the relative standard deviation (coefficient of variation) of the
        projected time-of-flight above which values are masked.
    """
    # Finally, mask regions with large uncertainty with NaNs.
    relative_error = sc.stddevs(table.data) / sc.values(table.data)
    mask = relative_error > sc.scalar(error_threshold)
    # Use numpy for indexing as table is 2D
    table.values[mask.values] = np.nan


def _compute_mean_tof_in_distance_range(
    simulation: SimulationResults,
    distance_bins: sc.Variable,
    time_bins: sc.Variable,
    distance_unit: str,
    time_unit: str,
    frame_period: sc.Variable,
    time_bins_half_width: sc.Variable,
) -> sc.DataArray:
    """
    Compute the mean time-of-flight inside event_time_offset bins for a given range of
    distances.

    Parameters
    ----------
    simulation:
        Results of a time-of-flight simulation used to create a lookup table.
    distance_bins:
        Bin edges for the distance axis in the lookup table.
    time_bins:
        Bin edges for the event_time_offset axis in the lookup table.
    distance_unit:
        Unit of the distance axis.
    time_unit:
        Unit of the event_time_offset axis.
    frame_period:
        Period of the source pulses, i.e., time between consecutive pulse starts.
    time_bins_half_width:
        Half width of the time bins in the event_time_offset axis.
    """
    simulation_distance = simulation.distance.to(unit=distance_unit)
    distances = sc.midpoints(distance_bins)
    # Compute arrival and flight times for all neutrons
    toas = simulation.time_of_arrival + (distances / simulation.speed).to(
        unit=time_unit, copy=False
    )
    dist = distances + simulation_distance
    tofs = dist * (sc.constants.m_n / sc.constants.h) * simulation.wavelength

    data = sc.DataArray(
        data=sc.broadcast(simulation.weight, sizes=toas.sizes),
        coords={
            "toa": toas,
            "tof": tofs.to(unit=time_unit, copy=False),
            "distance": dist,
        },
    ).flatten(to="event")

    # Add the event_time_offset coordinate, wrapped to the frame_period
    data.coords['event_time_offset'] = data.coords['toa'] % frame_period

    # Because we staggered the mesh by half a bin width, we want the values above
    # the last bin edge to wrap around to the first bin.
    # Technically, those values should end up between -0.5*bin_width and 0, but
    # a simple modulo also works here because even if they end up between 0 and
    # 0.5*bin_width, we are (below) computing the mean between -0.5*bin_width and
    # 0.5*bin_width and it yields the same result.
    # data.coords['event_time_offset'] %= pulse_period - time_bins_half_width
    data.coords['event_time_offset'] %= frame_period - time_bins_half_width

    binned = data.bin(
        distance=distance_bins + simulation_distance, event_time_offset=time_bins
    )

    # Weighted mean of tof inside each bin
    mean_tof = (
        binned.bins.data * binned.bins.coords["tof"]
    ).bins.sum() / binned.bins.sum()
    # Compute the variance of the tofs to track regions with large uncertainty
    variance = (
        binned.bins.data * (binned.bins.coords["tof"] - mean_tof) ** 2
    ).bins.sum() / binned.bins.sum()

    mean_tof.variances = variance.values
    return mean_tof


def make_tof_lookup_table(
    simulation: SimulationResults,
    ltotal_range: LtotalRange,
    distance_resolution: DistanceResolution,
    time_resolution: TimeResolution,
    pulse_period: PulsePeriod,
    pulse_stride: PulseStride,
    error_threshold: LookupTableRelativeErrorThreshold,
) -> TimeOfFlightLookupTable:
    """
    Compute a lookup table for time-of-flight as a function of distance and
    time-of-arrival.

    Parameters
    ----------
    simulation:
        Results of a time-of-flight simulation used to create a lookup table.
        The results should be a flat table with columns for time-of-arrival, speed,
        wavelength, and weight.
    ltotal_range:
        Range of total flight path lengths from the source to the detector.
    distance_resolution:
        Resolution of the distance axis in the lookup table.
    time_resolution:
        Resolution of the time-of-arrival axis in the lookup table. Must be an integer.
    pulse_period:
        Period of the source pulses, i.e., time between consecutive pulse starts.
    pulse_stride:
        Stride of used pulses. Usually 1, but may be a small integer when
        pulse-skipping.
    error_threshold:
        Threshold for the relative standard deviation (coefficient of variation) of the
        projected time-of-flight above which values are masked.

    Notes
    -----

    Below are some details about the binning and wrapping around frame period in the
    time dimension.

    We have some simulated ``toa`` (events) from a Tof/McStas simulation.
    Those are absolute ``toa``, unwrapped.
    First we compute the usual ``event_time_offset = toa % frame_period``.

    Now, we want to ensure periodic boundaries. If we make a bin centered around 0,
    and a bin centered around 71ms: the first bin will use events between 0 and
    ``0.5 * dt`` (where ``dt`` is the bin width).
    The last bin will use events between ``frame_period - 0.5*dt`` and
    ``frame_period + 0.5 * dt``. So when we compute the mean inside those two bins,
    they will not yield the same results.
    It is as if the first bin is missing the events it should have between
    ``-0.5 * dt`` and 0 (because of the modulo we computed above).

    To fix this, we do not make a last bin around 71ms (the bins stop at
    ``frame_period - 0.5*dt``). Instead, we compute modulo a second time,
    but this time using ``event_time_offset %= (frame_period - 0.5*dt)``.
    (we cannot directly do ``event_time_offset = toa % (frame_period - 0.5*dt)`` in a
    single step because it would introduce a gradual shift,
    as the pulse number increases).

    This second modulo effectively takes all the events that would have gone in the
    last bin (between ``frame_period - 0.5*dt`` and ``frame_period``) and puts them in
    the first bin. Instead of placing them between ``-0.5*dt`` and 0,
    it places them between 0 and ``0.5*dt``, but this does not really matter,
    because we then take the mean inside the first bin.
    Whether the events are on the left or right side of zero does not matter.

    Finally, we make a copy of the left edge, and append it to the right of the table,
    thus ensuring that the values on the right edge are strictly the same as on the
    left edge.
    """
    distance_unit = "m"
    time_unit = simulation.time_of_arrival.unit
    res = distance_resolution.to(unit=distance_unit)
    pulse_period = pulse_period.to(unit=time_unit)
    frame_period = pulse_period * pulse_stride

    min_dist, max_dist = (
        x.to(unit=distance_unit) - simulation.distance.to(unit=distance_unit)
        for x in ltotal_range
    )
    # We need to bin the data below, to compute the weighted mean of the wavelength.
    # This results in data with bin edges.
    # However, the 2d interpolator expects bin centers.
    # We want to give the 2d interpolator a table that covers the requested range,
    # hence we need to extend the range by at least half a resolution in each direction.
    # Then, we make the choice that the resolution in distance is the quantity that
    # should be preserved. Because the difference between min and max distance is
    # not necessarily an integer multiple of the resolution, we need to add a pad to
    # ensure that the last bin is not cut off. We want the upper edge to be higher than
    # the maximum distance, hence we pad with an additional 1.5 x resolution.
    pad = 2.0 * res
    distance_bins = sc.arange('distance', min_dist - pad, max_dist + pad, res)

    # Create some time bins for event_time_offset.
    # We want our final table to strictly cover the range [0, frame_period].
    # However, binning the data associates mean values inside the bins to the bin
    # centers. Instead, we stagger the mesh by half a bin width so we are computing
    # values for the final mesh edges (the bilinear interpolation needs values on the
    # edges/corners).
    nbins = int(frame_period / time_resolution.to(unit=time_unit)) + 1
    time_bins = sc.linspace(
        'event_time_offset', 0.0, frame_period.value, nbins + 1, unit=pulse_period.unit
    )
    time_bins_half_width = 0.5 * (time_bins[1] - time_bins[0])
    time_bins -= time_bins_half_width

    # To avoid a too large RAM usage, we compute the table in chunks, and piece them
    # together at the end.
    ndist = len(distance_bins) - 1
    max_size = 2e7
    total_size = ndist * len(simulation.time_of_arrival)
    nchunks = math.ceil(total_size / max_size)
    chunk_size = math.ceil(ndist / nchunks)
    pieces = []
    for i in range(nchunks):
        dist_edges = distance_bins[i * chunk_size : (i + 1) * chunk_size + 1]
        pieces.append(
            _compute_mean_tof_in_distance_range(
                simulation=simulation,
                distance_bins=dist_edges,
                time_bins=time_bins,
                distance_unit=distance_unit,
                time_unit=time_unit,
                frame_period=frame_period,
                time_bins_half_width=time_bins_half_width,
            )
        )

    table = sc.concat(pieces, 'distance')
    table.coords["distance"] = sc.midpoints(table.coords["distance"])
    table.coords["event_time_offset"] = sc.midpoints(table.coords["event_time_offset"])

    # Copy the left edge to the right to create periodic boundary conditions
    table = sc.DataArray(
        data=sc.concat(
            [table.data, table.data['event_time_offset', 0]], dim='event_time_offset'
        ),
        coords={
            "distance": table.coords["distance"],
            "event_time_offset": sc.concat(
                [table.coords["event_time_offset"], frame_period],
                dim='event_time_offset',
            ),
            "pulse_period": pulse_period,
            "pulse_stride": sc.scalar(pulse_stride, unit=None),
            "distance_resolution": table.coords["distance"][1]
            - table.coords["distance"][0],
            "time_resolution": table.coords["event_time_offset"][1]
            - table.coords["event_time_offset"][0],
            "error_threshold": sc.scalar(error_threshold),
        },
    )

    # In-place masking for better performance
    _mask_large_uncertainty(table, error_threshold)

    return TimeOfFlightLookupTable(table)


def simulate_chopper_cascade_using_tof(
    choppers: DiskChoppers,
    source_position: SourcePosition,
    neutrons: NumberOfSimulatedNeutrons,
    pulse_stride: PulseStride,
    seed: SimulationSeed,
    facility: SimulationFacility,
) -> SimulationResults:
    """
    Simulate a pulse of neutrons propagating through a chopper cascade using the
    ``tof`` package (https://tof.readthedocs.io).

    Parameters
    ----------
    choppers:
        A dict of DiskChopper objects representing the choppers in the beamline. See
        https://scipp.github.io/scippneutron/user-guide/chopper/processing-nexus-choppers.html#Build-DiskChopper
        for more information.
    source_position:
        A scalar variable with ``dtype=vector3`` that defines the source position.
        Must be in the same coordinate system as the choppers' axle positions.
    neutrons:
        Number of neutrons to simulate.
    pulse_stride:
        The pulse strinde; we need to simulate at least enough pulses to cover the
        requested stride.
    seed:
        Seed for the random number generator used in the simulation.
    facility:
        Facility where the experiment is performed.
    """
    import tof

    tof_choppers = [
        tof.Chopper(
            frequency=abs(ch.frequency),
            direction=tof.AntiClockwise
            if (ch.frequency.value > 0.0)
            else tof.Clockwise,
            open=ch.slit_begin,
            close=ch.slit_end,
            phase=abs(ch.phase),
            distance=sc.norm(
                ch.axle_position - source_position.to(unit=ch.axle_position.unit)
            ),
            name=name,
        )
        for name, ch in choppers.items()
    ]
    source = tof.Source(
        facility=facility, neutrons=neutrons, pulses=pulse_stride, seed=seed
    )
    if not tof_choppers:
        events = source.data.squeeze().flatten(to='event')
        return SimulationResults(
            time_of_arrival=events.coords["birth_time"],
            speed=events.coords["speed"],
            wavelength=events.coords["wavelength"],
            weight=events.data,
            distance=0.0 * sc.units.m,
        )
    model = tof.Model(source=source, choppers=tof_choppers)
    results = model.run()
    # Find name of the furthest chopper in tof_choppers
    furthest_chopper = max(tof_choppers, key=lambda c: c.distance)
    events = results[furthest_chopper.name].data.squeeze().flatten(to='event')
    events = events[
        ~(events.masks["blocked_by_others"] | events.masks["blocked_by_me"])
    ]
    return SimulationResults(
        time_of_arrival=events.coords["toa"],
        speed=events.coords["speed"],
        wavelength=events.coords["wavelength"],
        weight=events.data,
        distance=furthest_chopper.distance,
    )


def TofLookupTableWorkflow():
    """
    Create a workflow for computing a time-of-flight lookup table from a
    simulation of neutrons propagating through a chopper cascade.
    """
    wf = sl.Pipeline(
        (make_tof_lookup_table, simulate_chopper_cascade_using_tof),
        params={
            PulsePeriod: 1.0 / sc.scalar(14.0, unit="Hz"),
            PulseStride: 1,
            DistanceResolution: sc.scalar(0.1, unit="m"),
            TimeResolution: sc.scalar(250.0, unit='us'),
            LookupTableRelativeErrorThreshold: 0.1,
            NumberOfSimulatedNeutrons: 1_000_000,
            SimulationSeed: None,
            SimulationFacility: 'ess',
        },
    )
    return wf



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from typing import NewType

import sciline as sl
import scipp as sc

from ..nexus.types import MonitorType, RunType

TimeOfFlightLookupTableFilename = NewType("TimeOfFlightLookupTableFilename", str)
"""Filename of the time-of-flight lookup table."""


TimeOfFlightLookupTable = NewType("TimeOfFlightLookupTable", sc.DataArray)
"""
Lookup table giving time-of-flight as a function of distance and time of arrival.
"""


PulseStrideOffset = NewType("PulseStrideOffset", int | None)
"""
When pulse-skipping, the offset of the first pulse in the stride. This is typically
zero but can be a small integer < pulse_stride. If None, a guess is made.
"""


class DetectorLtotal(sl.Scope[RunType, sc.Variable], sc.Variable):
    """Total path length of neutrons from source to detector (L1 + L2)."""


class MonitorLtotal(sl.Scope[RunType, MonitorType, sc.Variable], sc.Variable):
    """Total path length of neutrons from source to monitor."""


class DetectorTofData(sl.Scope[RunType, sc.DataArray], sc.DataArray):
    """Detector data with time-of-flight coordinate."""


class MonitorTofData(sl.Scope[RunType, MonitorType, sc.DataArray], sc.DataArray):
    """Monitor data with time-of-flight coordinate."""



"""Adapter for loading NeXus files encoded as JSON."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import numpy as np
import scippnexus as snx

_nexus_class = "NX_class"
_nexus_units = "units"
_nexus_name = "name"
_nexus_path = "path"
_nexus_values = "values"
_nexus_dataset = "dataset"
_nexus_config = "config"
_nexus_group = "group"
_nexus_children = "children"
_nexus_link = "link"
_nexus_stream = "stream"

_filewriter_to_supported_numpy_dtype = {
    "float32": np.float32,
    "float64": np.float64,
    "float": np.float32,
    "double": np.float64,
    "int8": np.int32,
    "int16": np.int32,
    "int32": np.int32,
    "int64": np.int64,
    "uint8": np.int32,
    "uint16": np.int32,
    "uint32": np.int32,
    "uint64": np.int64,
    "string": np.str_,
}

numpy_to_filewriter_type = {
    np.float32: "float32",
    np.float64: "float64",
    np.int8: "int8",
    np.int16: "int16",
    np.int32: "int32",
    np.int64: "int64",
    np.uint8: "uint8",
    np.uint16: "uint16",
    np.uint32: "uint32",
    np.uint64: "uint64",
    np.str_: "string",
    np.object_: "string",
}


def json_nexus_group(
    json_dict: dict[str, Any], *, definitions: dict[str, type] | None = None
) -> snx.Group:
    """Parse a JSON dictionary into a NeXus group.

    Parameters
    ----------
    json_dict:
        ``dict`` containing a NeXus structure as JSON.
    definitions:
        ScippNexus application definitions.
        When not given, the default definitions are used.

    Returns
    -------
    :
        A NeXus group that can be used for loading data as if it were
        loaded from a file with :class:`scippnexus.File`.
    """
    return snx.Group(
        JSONGroup(json_dict),
        definitions=definitions if definitions is not None else snx.base_definitions(),
    )


class MissingAttribute(Exception):
    pass


def make_json_attr(name: str, value) -> dict:
    if isinstance(value, str | bytes):
        attr_info = {"string_size": len(value), "type": "string"}
    elif isinstance(value, float):
        attr_info = {"size": 1, "type": "float64"}
    elif isinstance(value, int):
        attr_info = {"size": 1, "type": "int64"}
    elif isinstance(value, list):
        attr_info = {"size": len(value), "type": "string"}
    else:
        attr_info = {
            "size": value.shape,
            "type": numpy_to_filewriter_type[value.dtype.type],
        }
    name_and_value = {"name": name, "values": value}
    return {**attr_info, **name_and_value}


def make_json_dataset(name: str, data) -> dict:
    if isinstance(data, str | bytes):
        dataset_info = {"string_size": len(data), "type": "string"}
    elif isinstance(data, float):
        dataset_info = {"size": 1, "type": "float64"}
    elif isinstance(data, int):
        dataset_info = {"size": 1, "type": "int32"}
    else:
        dataset_info = {
            "size": data.shape,
            "type": numpy_to_filewriter_type[data.dtype.type],
        }
    return {
        'module': _nexus_dataset,
        _nexus_config: {
            **dataset_info,
            _nexus_name: name,
            _nexus_values: data,
        },
        "attributes": [],
    }


def _get_attribute_value(
    element: dict, attribute_name: str
) -> str | float | int | list:
    """
    attributes can be a dictionary of key-value pairs, or an array
    of dictionaries with key, value, type, etc
    """
    try:
        attributes = element["attributes"]
        try:
            return attributes[attribute_name]
        except TypeError:
            for attribute in attributes:
                if attribute[_nexus_name] == attribute_name:
                    return attribute[_nexus_values]
    except KeyError:
        pass
    raise MissingAttribute


def _visitnodes(root: dict):
    for child in root.get(_nexus_children, ()):
        yield child
        yield from _visitnodes(child)


def _name(node: dict):
    if _nexus_name in node:
        return node[_nexus_name]
    if _nexus_config in node:
        return node[_nexus_config][_nexus_name]
    return ''


def _is_group(node: dict):
    return _nexus_children in node


def _is_dataset(node: dict):
    return node.get('module') == _nexus_dataset


def _is_link(node: dict):
    return node.get('module') == _nexus_link


def _is_stream(node: dict):
    return 'module' in node and not (_is_dataset(node) or _is_link(node))


def contains_stream(group: JSONGroup) -> bool:
    """Return True if the group contains a stream object"""
    return (
        isinstance(group, JSONGroup)
        and _nexus_children in group._node
        and any(map(_is_stream, group._node[_nexus_children]))
    )


class JSONTypeStringID:
    def get_cset(self):
        import h5py

        return h5py.h5t.CSET_UTF8


class JSONAttrID:
    def __init__(self):
        pass

    def get_type(self):
        return JSONTypeStringID()


class JSONAttributeManager(Mapping[str, Any]):
    def __init__(self, node: dict):
        self._node = node

    def __contains__(self, name):
        try:
            self[name]
        except MissingAttribute:
            return False
        return True

    def __getitem__(self, name):
        return _get_attribute_value(self._node, name)

    def __setitem__(self, name, value):
        if name in self:
            raise NotImplementedError("Replacing existing item not implemented yet.")
        attr = make_json_attr(name, value)
        self._node['attributes'].append(attr)
        return self[name]

    def __iter__(self):
        if (attrs := self._node.get('attributes')) is not None:
            if isinstance(attrs, dict):
                yield from attrs
            else:
                for item in attrs:
                    yield item[_nexus_name]

    def __len__(self):
        return sum(1 for _ in self)

    def get(self, name: str, default=None):
        return self[name] if name in self else default

    def get_id(self, name) -> JSONAttrID:
        # TODO This is a hack that only works since this is used only for a single
        # purpose by scippnexus.NXobject
        return JSONAttrID()


class JSONNode:
    def __init__(self, node: dict, *, parent=None):
        self._file = parent.file if parent is not None else self
        self._parent = self if parent is None else parent
        self._node = node
        name = _name(self._node)
        if parent is None or parent.name == '/':
            self._name = f'/{name}'
        else:
            self._name = f'{parent.name}/{name}'

    @property
    def attrs(self) -> JSONAttributeManager:
        return JSONAttributeManager(self._node)

    @property
    def name(self) -> str:
        return self._name

    @property
    def file(self):
        return self._file

    @property
    def parent(self):
        return self._parent


class JSONDataset(JSONNode):
    @property
    def dtype(self) -> str:
        try:
            dtype = self._node[_nexus_config]["type"]
        except KeyError:
            if "dtype" not in self._node[_nexus_config] and isinstance(
                self._node[_nexus_config][_nexus_values], str
            ):
                dtype = 'string'
            else:
                dtype = self._node[_nexus_config]["dtype"]
        if dtype == 'string':
            return np.dtype(str)
        return np.dtype(dtype)

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def shape(self):
        return np.asarray(self._node[_nexus_config][_nexus_values]).shape

    def __getitem__(self, index):
        return np.asarray(self._node[_nexus_config][_nexus_values])[index]

    def read_direct(self, buf, source_sel):
        buf[...] = self[source_sel]

    def asstr(self, **ignored):
        return self


class JSONGroup(JSONNode):
    def __contains__(self, name: str) -> bool:
        try:
            self[name]
            return True
        except KeyError:
            return False

    def keys(self) -> list[str]:
        if contains_stream(self):
            return []
        children = self._node[_nexus_children]
        return [_name(child) for child in children if not contains_stream(child)]

    def items(self) -> list[tuple[str, JSONNode]]:
        return [(key, self[key]) for key in self.keys()]

    def _as_group_or_dataset(self, item, parent):
        if _is_group(item):
            return JSONGroup(item, parent=parent)
        return JSONDataset(item, parent=parent)

    def __getitem__(self, name: str) -> JSONDataset | JSONGroup:
        if name.startswith('/') and name.count('/') == 1:
            parent = self.file
        elif '/' in name:
            parent = self['/'.join(name.split('/')[:-1])]
        else:
            parent = self

        for child in parent._node[_nexus_children]:
            if _name(child) != name.split('/')[-1]:
                continue
            if _is_link(child):
                return self[child[_nexus_config]["target"]]
            if _is_group(child) or _is_dataset(child):
                return self._as_group_or_dataset(child, parent)

        raise KeyError(f"Unable to open object (object '{name}' doesn't exist)")

    def __iter__(self):
        yield from self.keys()

    def visititems(self, callable):
        def skip(node):
            return _is_link(node) or contains_stream(self)

        children = [
            _name(child) for child in self._node[_nexus_children] if not skip(child)
        ]
        for key in children:
            item = self[key]
            callable(key, item)
            if isinstance(item, JSONGroup):
                item.visititems(callable)

    def create_dataset(self, name: str, data) -> JSONDataset:
        if name in self:
            raise NotImplementedError("Replacing existing item not implemented yet.")
        dataset = make_json_dataset(name, data)
        self._node[_nexus_children].append(dataset)
        return self[name]

    def create_group(self, name: str) -> JSONGroup:
        if name in self:
            raise NotImplementedError("Replacing existing item not implemented yet.")
        group = {"type": "group", "name": name, "children": [], "attributes": []}
        self._node[_nexus_children].append(group)
        return self[name]
